#ifndef PYROTRASH_H
#define PYROTRASH_H

//main.cpp

#include "datafile.h"

extern DATAFILE *dat;

extern BITMAP *scrbuf;

#define TEST_ONE_PLAYER false

//game.cpp

#define MAP_W 20
#define MAP_H 13

#define TILE_W 32
#define TILE_H 32

//Four players, two miners
//This variable is limited to 12 (8 miners) because keys for miners are stored in an int
#define NUMCHARS 12

#define MAXHEALTH 10000
#define MINERHEALTH 1000

#define K_LEFT 1
#define K_RIGHT 2
#define K_UP 4
#define K_DOWN 8
#define K_SHOOT 16
#define K_SERVER_ENFORCED 32
#define MINER_KSHIFT 4
#define PLAYER_KSHIFT 6

//Note: some code in getMinerKeys() depends on the directionals being 0-3
#define MT_LEFT 0
#define MT_RIGHT 1
#define MT_UP 2
#define MT_DOWN 3
#define MT_DIE 4
//the following one is only used for players, not for shots
#define MT_WIN 5

//fire points per tile
#define FIREGRAN 16
//maximum value each fire point can reach
#define FIREMAX 1023
//this takes the fire values down to the 0-127 range used by fire.cpp
#define FIRE_TO_BITMAP_SHIFT 3
//value a fire point has to reach before it spreads
#define FIRETHRESH 50

extern bool walkable(int t);

struct Game;

struct Character {
	bool active;
	int x,y;
	int xv,yv;
	int health;
	int shootdelay;
	int movetype,moveframe;
	int shootframe;
	//AI state for miners
	int lastchangex,lastchangey;
	int wantedmovetype;
	//Customisation
	int width,height;
	int runaccel,runspeed;
	int climbaccel,climbspeed;
	int fallspeed;
	int shootinterval,shotspeed,shotdamage;
	int firedefence;
	//Animation stuff
	int moveframebase[6],moveframeloop[6],moveframecount[6],moveframetime[6];
	int shootframebase[4],shootframecount,shootframetime;
	int shotframebase[5],shotframecount[5],shotframetime[5];
	
	void move(Game *g, int id, int keys, int ticksinadvance, bool dead);//used for prospective moving
	void animate(Game *g, int id);//not used for prospective moving
	void draw();
};

struct Shot {
	Shot *next;
	int id;//who shot it
	int x,y;
	int xv,yv;
	int movetype;
	int frame;
	//used on client - when it hits me, I'm actually being drawn having moved
	//so this is set, and the shot is kept around and immediately goes for me
	bool hitme;
	
	void draw(Game *g);
};

struct Game {
	unsigned char map[MAP_H][MAP_W];
	unsigned short hfire[MAP_H+1][MAP_W*FIREGRAN+2];
	unsigned short vfire[MAP_W+1][MAP_H*FIREGRAN+2];
	Character chr[NUMCHARS];
	int time;
	int survivaltime[4];
	bool finished;
	Shot *shots;
	
	Game();
	~Game();
	void generateMap();
	void placePlayer(int i);
	void placePlayer(int i, int x, int y, int movetype);
	void placeMiners();
	int getMinerKeys();
	void update(int playerkeys, int minerkeys, int myid, int projx, int projy);
	void updateEnding(int myid, int projx, int projy);
	void draw(int myid);
	
	private:
	void updateShots(int myid, int projx, int projy);
};

//fire.cpp

#define WARPGRAN 2
#define WARP_W (MAP_W*WARPGRAN+1)
#define WARP_H (MAP_H*WARPGRAN+1)
#define WARP_XD (FIREGRAN/WARPGRAN)
#define WARP_YD (FIREGRAN/WARPGRAN)
#define FIRE_W (MAP_W*FIREGRAN)
#define FIRE_H (MAP_H*FIREGRAN)

struct WarpPoint {
	int x,y;
	int dx,dy;
};

struct Fire {
	static const int UPWIND=8192;
	
	WarpPoint warpU[WARP_H][WARP_W];//unconstrained
	WarpPoint warpC[WARP_H][WARP_W];//constrained
	//These have room for blur overrun at all edges + UPWIND read overrun at bottom
	unsigned char fire1[FIRE_H+2+UPWIND][FIRE_W+2];
	unsigned char fire2[FIRE_H+2+UPWIND][FIRE_W+2];
	unsigned char (*src)[FIRE_W+2];
	unsigned char (*dst)[FIRE_W+2];
	int tick;
	
	static void init();
	Fire();
	~Fire();
	void update(Game *g);
	void draw();
	
	private:
	void constrainWarpToMap(Game *g);
	void addFireFromGame(Game *g);
};

#endif
