#include <cstdlib>
#include <cstring>
#include <allegro.h>
#include "pyrotrash.h"

using namespace std;

static WarpPoint pnt;
static WarpPoint row[WARP_W];
static unsigned char reduce[128*8];//allows fire values up to 127
static unsigned char remap[240][128*2];//allows fire values up to 127

void Fire::init() {
	for (int i=0; i<128*8; i++) reduce[i]=i*15/128;
	remap[0][0]=0;
	for (int j=1; j<128*2; j++) remap[0][j]=240+(j>>4);
	for (int i=1; i<240; i++) {
		for (int j=0; j<128; j++) remap[i][j]=i;
		for (int j=128; j<128*2; j++) remap[i][j]=240+(j>>4);
	}
}

Fire::Fire() {
	//Randomise warp map
	for (int y=0; y<WARP_H; y++) {
		for (int x=0; x<WARP_W; x++) {
			warpU[y][x].x=(rand()*16+rand()%16)%131073-65536;
			warpU[y][x].y=(rand()*16+rand()%16)%131073-65536+UPWIND;
			warpU[y][x].dx=0;
			warpU[y][x].dy=0;
		}
	}
	//Clamp some values
	for (int y=0; y<WARP_H; y++) warpU[y][0].x=0;
	for (int y=0; y<WARP_H; y++) warpU[y][WARP_W-1].x=0;
	for (int x=0; x<WARP_W; x++) warpU[0][x].y=UPWIND;
	for (int x=0; x<WARP_W; x++) warpU[WARP_H-1][x].y=UPWIND;
	//Clear bitmaps
	memset(fire1,0,sizeof fire1);
	memset(fire2,0,sizeof fire2);
	//Set pointers up
	src=fire1; dst=fire2;
	//Start multi-step phasing
	tick=0;
}

Fire::~Fire() {
}

#define LIMIT(v,min,max) \
	v-=(min); v&=(signed)~v>>31; v+=(min); \
	v=(max)-v; v&=(signed)~v>>31; v=(max)-v

#define SIGN(v) (((signed)(v)>>31)+((unsigned)-(v)>>31))

void Fire::update(Game *g) {
	if (tick==0) {
		//Update warp point velocities - pull towards nearby one but also towards 0,UPWIND position
		for (int y=1; y<WARP_H-1; y++) {
			for (int x=1; x<WARP_W-1; x++) {
				warpU[y][x].dx+=(warpU[y][x+1].x+warpU[y][x-1].x-(warpU[y][x].x<<2)+128)>>8;
				warpU[y][x].dy+=(warpU[y+1][x].y+warpU[y-1][x].y-(warpU[y][x].y<<2)+(UPWIND<<1)+128)>>8;
				//Make sure it keeps moving, but limit it
				int v,w;
				v=warpU[y][x].dx; w=v/512; warpU[y][x].dx+=SIGN(v)-(SIGN(w)<<1);
				v=warpU[y][x].dy; w=v/512; warpU[y][x].dy+=SIGN(v)-(SIGN(w)<<1);
			}
		}
		//Update warp point positions
		for (int y=1; y<WARP_H-1; y++) {
			for (int x=1; x<WARP_W-1; x++) {
				int v;
				v=warpU[y][x].x+warpU[y][x].dx; LIMIT(v,-65536,65536); warpU[y][x].x=v;
				v=warpU[y][x].y+warpU[y][x].dy; LIMIT(v,UPWIND-65536,UPWIND+65536); warpU[y][x].y=v;
			}
		}
		//Constrain some warp points depending on the level
		constrainWarpToMap(g);
		//Initialise row[] to the top of the warp array
		for (int xw=0; xw<WARP_W; xw++) {
			row[xw].x=warpC[0][xw].x;
			row[xw].y=warpC[0][xw].y;
		}
	}
	//For each row of the warp array - do this over four ticks
	int ywend=(tick+1)*(WARP_H-1)/4;
	for (int yw=tick*(WARP_H-1)/4,ys=yw*WARP_YD,ye=ys+WARP_YD; yw<ywend; yw++,ys=ye,ye+=WARP_YD) {
		//Prepare to interpolate row[]
		for (int xw=0; xw<WARP_W; xw++) {
			row[xw].dx=(warpC[yw+1][xw].x-row[xw].x)/WARP_YD;
			row[xw].dy=(warpC[yw+1][xw].y-row[xw].y)/WARP_YD;
		}
		//For each row of pixels
		for (int y=ys; y<ye; y++) {
			//Initialise pnt to the left of row[]
			pnt.x=row[0].x;
			pnt.y=row[0].y;
			//For each point in row[]
			for (int xw=0,xs=0,xe=WARP_XD; xw<WARP_W-1; xw++,xs=xe,xe+=WARP_XD) {
				//Prepare to interpolate pnt
				pnt.dx=(row[xw+1].x-pnt.x)/WARP_XD;
				pnt.dy=(row[xw+1].y-pnt.y)/WARP_XD;
				//For each pixel
				unsigned char *d=dst[1+y]+1;
				for (int x=xs; x<xe; x++) {
					//Apply the warp/blur effect - offset by up to four pixels at once
					int ysrc=1+y+((pnt.y+8192)>>14);
					int xsrc=1+x+((pnt.x+8192)>>14);
					d[x]=reduce[(src[ysrc][xsrc]<<2)+
					             src[ysrc][xsrc-1]+
					             src[ysrc][xsrc+1]+
					             src[ysrc-1][xsrc]+
					             src[ysrc+1][xsrc]];
					//Interpolate pnt
					pnt.x+=pnt.dx;
					pnt.y+=pnt.dy;
				}
			}
			//Interpolate row[]
			for (int xw=0; xw<WARP_W; xw++) {
				row[xw].x+=row[xw].dx;
				row[xw].y+=row[xw].dy;
			}
		}
	}
	//Swap the bitmaps, and add new fire to the new bitmap
	if (tick==3) {
		unsigned char (*tmp)[FIRE_W+2]=src; src=dst; dst=tmp;
		addFireFromGame(g);
	}
	//Next tick
	tick=(tick+1)&3;
}

void Fire::draw() {
	for (int y=0; y<FIRE_H; y++) {
		unsigned char *srctop=src[1+y]+1;
		unsigned char *srcbottom=src[1+y+1]+1;
		unsigned char *top=scrbuf->line[y*2];
		unsigned char *bottom=scrbuf->line[y*2+1];
		for (int x=0; x<FIRE_W; x++) {
			int ftl=*srctop++;
			int fbl=*srcbottom++;
			int ftr=ftl+*srctop;
			int fbr=ftr+fbl+*srcbottom;
			fbl+=ftl;
			*top=remap[*top][ftl<<1]; top++;
			*top=remap[*top][ftr]; top++;
			*bottom=remap[*bottom][fbl]; bottom++;
			*bottom=remap[*bottom][fbr>>1]; bottom++;
		}
	}
}

void Fire::constrainWarpToMap(Game *g) {
	memcpy(warpC,warpU,sizeof warpU);
	for (int y=0; y<MAP_H; y++) {
		for (int x=0; x<MAP_W; x++) {
			if (!walkable(g->map[y][x])) {
				if (warpC[y*WARPGRAN][x*WARPGRAN].x<0) warpC[y*WARPGRAN][x*WARPGRAN].x=0;
				if (warpC[y*WARPGRAN][x*WARPGRAN].y<0) warpC[y*WARPGRAN][x*WARPGRAN].y=0;
				if (warpC[y*WARPGRAN][(x+1)*WARPGRAN].x>0) warpC[y*WARPGRAN][(x+1)*WARPGRAN].x=0;
				if (warpC[y*WARPGRAN][(x+1)*WARPGRAN].y<0) warpC[y*WARPGRAN][(x+1)*WARPGRAN].y=0;
				if (warpC[(y+1)*WARPGRAN][x*WARPGRAN].x<0) warpC[(y+1)*WARPGRAN][x*WARPGRAN].x=0;
				if (warpC[(y+1)*WARPGRAN][x*WARPGRAN].y>0) warpC[(y+1)*WARPGRAN][x*WARPGRAN].y=0;
				if (warpC[(y+1)*WARPGRAN][(x+1)*WARPGRAN].x>0) warpC[(y+1)*WARPGRAN][(x+1)*WARPGRAN].x=0;
				if (warpC[(y+1)*WARPGRAN][(x+1)*WARPGRAN].y>0) warpC[(y+1)*WARPGRAN][(x+1)*WARPGRAN].y=0;
			}
		}
	}
}

void Fire::addFireFromGame(Game *g) {
	for (int y=1; y<MAP_H; y++) {
		unsigned short *hfire=g->hfire[y];
		unsigned char *row1=src[y*FIREGRAN-2];
		unsigned char *row2=src[y*FIREGRAN-1];
		unsigned char *row3=src[y*FIREGRAN];
		unsigned char *row4=src[y*FIREGRAN+1];
		for (int x=0; x<MAP_W*FIREGRAN; x++) {
			int v=hfire[x+1]>>FIRE_TO_BITMAP_SHIFT;
			if (v>row2[x]) row2[x]=v;
			if (v>row3[x]) row3[x]=v;
			v-=48;
			if (v>row1[x]) row1[x]=v;
			if (v>row4[x]) row4[x]=v;
		}
	}
	for (int x=1; x<MAP_W; x++) {
		unsigned short *vfire=g->vfire[x];
		unsigned char (*srcoffset)[FIRE_W+2]=src;
		int off=x*FIREGRAN;
		for (int y=0; y<MAP_H*FIREGRAN; y++) {
			int v=vfire[y+1]>>FIRE_TO_BITMAP_SHIFT;
			if (v>srcoffset[y][off-1]) srcoffset[y][off-1]=v;
			if (v>srcoffset[y][off]) srcoffset[y][off]=v;
			v-=48;
			if (v>srcoffset[y][off-2]) srcoffset[y][off-2]=v;
			if (v>srcoffset[y][off+1]) srcoffset[y][off+1]=v;
		}
	}
}
