#define ALLEGRO_USE_CONSOLE

#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <allegro.h>
#include "pyrotrash.h"

#define T_CLEAR 0
#define T_LADDER 1
#define T_ROCK 2
#define T_COAL 3
#define T_TL_COAL 4
#define T_TR_COAL 5
#define T_BL_COAL 6
#define T_BR_COAL 7

#define FALLACCEL 2

using namespace std;

bool walkable(int t) {return t==T_CLEAR || t==T_LADDER;}
static bool iscoalleft(int t)   {return t==T_COAL || t==T_TL_COAL || t==T_BL_COAL;}
static bool iscoalright(int t)  {return t==T_COAL || t==T_TR_COAL || t==T_BR_COAL;}
static bool iscoaltop(int t)    {return t==T_COAL || t==T_TL_COAL || t==T_TR_COAL;}
static bool iscoalbottom(int t) {return t==T_COAL || t==T_BL_COAL || t==T_BR_COAL;}

#define SETMOVEFRAMES(id,dir,base,loop,count,time) \
	chr[id].moveframebase[dir]=base; \
	chr[id].moveframeloop[dir]=loop; \
	chr[id].moveframecount[dir]=count; \
	chr[id].moveframetime[dir]=time

#define SETSHOOTFRAMES(id,base,count,time) \
	for (int i=0; i<4; i++) chr[id].shootframebase[i]=base+count*i; \
	chr[id].shootframecount=count; \
	chr[id].shootframetime=time

#define SETSHOTFRAMES(id,base,count,time,diebase,diecount,dietime) \
	for (int i=0; i<4; i++) { \
		chr[id].shotframebase[i]=base; \
		chr[id].shotframecount[i]=count; \
		chr[id].shotframetime[i]=time; \
	} \
	chr[id].shotframebase[4]=diebase; \
	chr[id].shotframecount[4]=diecount; \
	chr[id].shotframetime[4]=dietime

Game::Game() {
	memset(hfire,0,sizeof(hfire));
	memset(vfire,0,sizeof(vfire));
	for (int i=0; i<NUMCHARS; i++) chr[i].active=false;
	time=0;
	for (int i=0; i<4; i++) survivaltime[i]=0x7FFFFFFF;
	finished=false;
	shots=NULL;
	//Crisp packet
	chr[0].width=192; chr[0].height=192;
	chr[0].runaccel=1; chr[0].runspeed=8;
	chr[0].climbaccel=1; chr[0].climbspeed=4;
	chr[0].fallspeed=8;
	chr[0].shootinterval=64; chr[0].shotspeed=16; chr[0].shotdamage=300;
	chr[0].firedefence=2048;
	SETMOVEFRAMES(0,MT_LEFT,CRISPLR_000,0,3,8);
	SETMOVEFRAMES(0,MT_RIGHT,CRISPLR_003,0,3,8);
	SETMOVEFRAMES(0,MT_UP,CRISPUD_000,0,4,8);
	SETMOVEFRAMES(0,MT_DOWN,CRISPUD_004,0,4,8);
	SETMOVEFRAMES(0,MT_DIE,CRISPDIE_000,3,4,12);
	SETMOVEFRAMES(0,MT_WIN,CRISPWIN_000,0,4,12);
	SETSHOOTFRAMES(0,CRISPSHOOT_000,3,8);
	SETSHOTFRAMES(0,CRISPSHOT_000,1,8,CRISPSHOTDIE_000,3,12);
	//Milk bottle
	chr[1].width=128; chr[1].height=192;
	chr[1].runaccel=1; chr[1].runspeed=12;
	chr[1].climbaccel=1; chr[1].climbspeed=8;
	chr[1].fallspeed=48;
	chr[1].shootinterval=48; chr[1].shotspeed=20; chr[1].shotdamage=200;
	chr[1].firedefence=2048;
	SETMOVEFRAMES(1,MT_LEFT,MILKMOVE_000,1,5,8);
	SETMOVEFRAMES(1,MT_RIGHT,MILKMOVE_005,1,5,8);
	SETMOVEFRAMES(1,MT_UP,MILKMOVE_010,1,5,8);
	SETMOVEFRAMES(1,MT_DOWN,MILKMOVE_010,1,5,8);
	SETMOVEFRAMES(1,MT_DIE,MILKDIE_000,4,5,12);
	SETMOVEFRAMES(1,MT_WIN,MILKWIN_000,0,4,12);
	SETSHOOTFRAMES(1,MILKSHOOT_000,3,12);
	SETSHOTFRAMES(1,MILKSHOT_000,1,8,MILKSHOTDIE_000,4,8);
	//Energy drink
	chr[2].width=192; chr[2].height=192;
	chr[2].runaccel=2; chr[2].runspeed=16;
	chr[2].climbaccel=2; chr[2].climbspeed=12;
	chr[2].fallspeed=64;
	chr[2].shootinterval=32; chr[2].shotspeed=32; chr[2].shotdamage=100;
	chr[2].firedefence=8192;
	SETMOVEFRAMES(2,MT_LEFT,CANMOVE_000,0,6,6);
	SETMOVEFRAMES(2,MT_RIGHT,CANMOVE_006,0,6,6);
	SETMOVEFRAMES(2,MT_UP,CANMOVE_012,0,6,6);
	SETMOVEFRAMES(2,MT_DOWN,CANMOVE_018,0,6,6);
	SETMOVEFRAMES(2,MT_DIE,CANDIE_000,5,6,12);
	SETMOVEFRAMES(2,MT_WIN,CANWIN_000,0,4,6);
	SETSHOOTFRAMES(2,CANSHOOT_000,2,8);
	SETSHOTFRAMES(2,CANSHOT_000,4,8,CANSHOTDIE_000,4,12);
	//Burnt chicken
	chr[3].width=192; chr[3].height=192;
	chr[3].runaccel=1; chr[3].runspeed=5;
	chr[3].climbaccel=1; chr[3].climbspeed=3;
	chr[3].fallspeed=64;
	chr[3].shootinterval=64; chr[3].shotspeed=24; chr[3].shotdamage=400;
	chr[3].firedefence=8192;
	SETMOVEFRAMES(3,MT_LEFT,CHICKENMOVE_000,1,5,8);
	SETMOVEFRAMES(3,MT_RIGHT,CHICKENMOVE_005,1,5,8);
	SETMOVEFRAMES(3,MT_UP,CHICKENMOVE_010,1,5,8);
	SETMOVEFRAMES(3,MT_DOWN,CHICKENMOVE_010,1,5,8);
	SETMOVEFRAMES(3,MT_DIE,CHICKENDIE_000,7,8,8);
	SETMOVEFRAMES(3,MT_WIN,CHICKENWIN_000,0,4,16);
	SETSHOOTFRAMES(3,CHICKENSHOOT_000,2,16);
	SETSHOTFRAMES(3,CHICKENSHOT_000,4,8,CHICKENSHOTDIE_000,4,8);
	//Miners
	for (int i=4; i<NUMCHARS; i++) {
		chr[i].width=192; chr[i].height=224;
		chr[i].runaccel=1; chr[i].runspeed=7;
		chr[i].climbaccel=1; chr[i].climbspeed=4;
		chr[i].fallspeed=32;
		chr[i].firedefence=1024;
		SETMOVEFRAMES(i,MT_LEFT,MINERMOVE_000,1,5,16);
		SETMOVEFRAMES(i,MT_RIGHT,MINERMOVE_005,1,5,16);
		SETMOVEFRAMES(i,MT_UP,MINERMOVE_010,1,5,16);
		SETMOVEFRAMES(i,MT_DOWN,MINERMOVE_010,1,5,16);
		SETMOVEFRAMES(i,MT_DIE,MINERDIE_000,4,5,12);
	}
}

Game::~Game() {
	while (shots!=NULL) {
		Shot *next=shots->next;
		delete shots;
		shots=next;
	}
}

void Game::generateMap() {
	//Create strata of rock and coal - mostly coal
	unsigned char strata[16];
	for (int i=0; i<16; i++) strata[i]=T_COAL;
	for (int i=0; i<3; i++) strata[(rand()>>8)&15]=T_ROCK;
	//Choose a gradient, -16 to 16
	int grad=rand()%33-16;
	for (int y=0; y<MAP_H; y++) {
		for (int x=0; x<MAP_W; x++) {
			//Calculate pos in strata array for centre x
			int pos_centre=8+(y-MAP_H/2);
			//Get val from strata array but take x into account
			//- For positive gradient, we use lower strata for greater x - so strata are /-shaped
			int val=strata[(pos_centre+(x-MAP_W/2)*grad/16)&15];
			//Get val for x+1
			int valtoright=strata[(pos_centre+(x+1-MAP_W/2)*grad/16)&15];
			//Use slanting tile if val to right is different
			if (valtoright!=val) {
				if (valtoright==T_COAL && grad>0) val=T_BR_COAL;
				if (valtoright==T_COAL && grad<0) val=T_TR_COAL;
				if (valtoright==T_ROCK && grad>0) val=T_TL_COAL;
				if (valtoright==T_ROCK && grad<0) val=T_BL_COAL;
			}
			map[y][x]=val;
		}
	}
	//Create ladders
	int mintop=1;
	while (true) {
		int minbottom=mintop+2+(rand()/(RAND_MAX/2))*2;
		while (minbottom>MAP_H-2) minbottom-=2;
		if (minbottom<=mintop) break;
		int top=mintop; while (top-2>=1 && rand()<RAND_MAX/4) top-=2;
		int bottom=minbottom; while (bottom+2<=MAP_H-2 && rand()<RAND_MAX/4) bottom+=2;
		tryagain:
		int x=1+rand()%(MAP_W-2);
		for (int y=top; y<=bottom; y++) {
			if (map[y][x-1]==T_LADDER || map[y][x+1]==T_LADDER) goto tryagain;
		}
		for (int y=top; y<=bottom; y++) map[y][x]=T_LADDER;
		mintop=minbottom;
	}
	//Create horizontal corridors
	const int minleft=MAP_W/3;
	const int minright=MAP_W*2/3;
	for (int y=1; y<MAP_H-1; y+=2) {
		int left=1,right=MAP_W-2;
		if (rand()<RAND_MAX/2) {while (left<minleft && map[y][left]!=T_LADDER) left++;}
		if (rand()<RAND_MAX/2) {while (right>minright && map[y][right]!=T_LADDER) right--;}
		for (int x=left; x<=right; x++) {if (map[y][x]!=T_LADDER) map[y][x]=T_CLEAR;}
	}
	//Create entrance
	while (true) {
		int x=minleft+rand()%(minright+1-minleft);
		if (map[1][x]!=T_LADDER) {
			map[0][x]=T_CLEAR;
			break;
		}
	}
}

void Game::placePlayer(int i) {
	int count=0;
	for (int y=1; y<MAP_H-1; y++) {
		for (int x=1; x<MAP_W-1; x++) {
			if (walkable(map[y][x]) && !walkable(map[y+1][x])) count++;
		}
	}
	int choice=rand()%count;
	count=0;
	for (int y=1; y<MAP_H-1; y++) {
		for (int x=1; x<MAP_W-1; x++) {
			if (walkable(map[y][x]) && !walkable(map[y+1][x])) {
				if (count==choice) {
					placePlayer(i,x*256+128-chr[i].width/2,y*256+256-chr[i].height/2,rand()<RAND_MAX/2?MT_LEFT:MT_RIGHT);
					return;
				}
				count++;
			}
		}
	}
}

void Game::placePlayer(int i, int x, int y, int movetype) {
	chr[i].active=true;
	chr[i].x=x; chr[i].y=y;
	chr[i].xv=chr[i].yv=0;
	chr[i].health=(i>=4?MINERHEALTH:MAXHEALTH);
	chr[i].shootdelay=0;
	chr[i].movetype=movetype; chr[i].moveframe=0;
	chr[i].shootframe=-1;
}

void Game::placeMiners() {
	for (int i=4; i<NUMCHARS; i++) {
		placePlayer(i);
		chr[i].lastchangex=chr[i].lastchangey=-1;
		chr[i].wantedmovetype=chr[i].movetype;
	}
}

int Game::getMinerKeys() {
	int minerkeys=0;
	for (int i=4; i<NUMCHARS; i++) {
		//If we haven't already changed direction for this square ...
		int xsq=(chr[i].x+chr[i].width/2)>>8,ysq=(chr[i].y+chr[i].height-1)>>8;
		if (xsq!=chr[i].lastchangex || ysq!=chr[i].lastchangey) {
			//Check which ways we can go from where we are
			int cango=0;
			if (walkable(map[ysq][xsq-1]) && !walkable(map[ysq+1][xsq-1])) cango|=1<<MT_LEFT;
			if (walkable(map[ysq][xsq+1]) && !walkable(map[ysq+1][xsq+1])) cango|=1<<MT_RIGHT;
			if (map[ysq][xsq]==T_LADDER && map[ysq-1][xsq]==T_LADDER) cango|=1<<MT_UP;
			if (map[ysq][xsq]==T_LADDER && map[ysq+1][xsq]==T_LADDER) cango|=1<<MT_DOWN;
			//Remove the direction that's opposite the way we're already going
			//- unless it's the only way to go
			if (chr[i].wantedmovetype==MT_LEFT && cango!=1<<MT_RIGHT) cango&=~(1<<MT_RIGHT);
			if (chr[i].wantedmovetype==MT_RIGHT && cango!=1<<MT_LEFT) cango&=~(1<<MT_LEFT);
			if (chr[i].wantedmovetype==MT_UP && cango!=1<<MT_DOWN) cango&=~(1<<MT_DOWN);
			if (chr[i].wantedmovetype==MT_DOWN && cango!=1<<MT_UP) cango&=~(1<<MT_UP);
			//Choose a random new direction to try to go in
			int count=0;
			for (int j=0; j<4; j++) {if (cango&(1<<j)) count++;}
			if (count>0) {
				int choice=rand()%count;
				count=0;
				for (int j=0; j<4; j++) {
					if (cango&(1<<j)) {
						if (choice==count) {chr[i].wantedmovetype=j; break;}
						count++;
					}
				}
			}
			//Make a note that we've changed in this square
			chr[i].lastchangex=xsq; chr[i].lastchangey=ysq;
		}
		//Now try and move in the target direction
		int keys=0;
		if (chr[i].wantedmovetype==MT_LEFT || chr[i].wantedmovetype==MT_RIGHT) {
			if (chr[i].wantedmovetype==MT_RIGHT) keys|=K_RIGHT; else keys|=K_LEFT;
			if (chr[i].y<(ysq<<8)+256-chr[i].height) keys|=K_DOWN;
			if (chr[i].y>(ysq<<8)+256-chr[i].height) keys|=K_UP;
		}
		if (chr[i].wantedmovetype==MT_UP || chr[i].wantedmovetype==MT_DOWN) {
			if (chr[i].wantedmovetype==MT_DOWN) keys|=K_DOWN; else keys|=K_UP;
			if (chr[i].x<(xsq<<8)+128-chr[i].width/2) keys|=K_RIGHT;
			if (chr[i].x>(xsq<<8)+128-chr[i].width/2) keys|=K_LEFT;
		}
		//Done!
		minerkeys|=keys<<((i-4)*MINER_KSHIFT);
	}
	return minerkeys;
}

void Game::update(int playerkeys, int minerkeys, int myid, int projx, int projy) {
	//Increase fire
	for (int y=1; y<MAP_H; y++) {
		for (int x=1; x<1+MAP_W*FIREGRAN; x++) {
			if (hfire[y][x]>0 && hfire[y][x]<FIREMAX) hfire[y][x]++;
		}
	}
	for (int x=1; x<MAP_W; x++) {
		for (int y=1; y<1+MAP_H*FIREGRAN; y++) {
			if (vfire[x][y]>0 && vfire[x][y]<FIREMAX) vfire[x][y]++;
		}
	}
	//Spread fire wherever there's coal - consider tile xt,yt and the fire line above / left of it
	for (int yt=1; yt<MAP_H; yt++) {
		for (int xt=0; xt<MAP_W; xt++) {
			if (iscoalbottom(map[yt-1][xt]) || iscoaltop(map[yt][xt])) {
				if (walkable(map[yt-1][xt]) || walkable(map[yt][xt])) {
					int xp=1+xt*FIREGRAN;
					int xpend=xp+FIREGRAN;
					if (hfire[yt][xp]==0) {
						if (vfire[xt][yt*FIREGRAN]>=FIRETHRESH || vfire[xt][1+yt*FIREGRAN]>=FIRETHRESH) hfire[yt][xp]=1;
					}
					for (; xp<xpend; xp++) {
						if (hfire[yt][xp]==0) {
							if (hfire[yt][xp-1]>=FIRETHRESH || hfire[yt][xp+1]>=FIRETHRESH) hfire[yt][xp]=1;
						}
					}
					xp--;
					if (hfire[yt][xp]==0) {
						if (vfire[xt+1][yt*FIREGRAN]>=FIRETHRESH || vfire[xt+1][1+yt*FIREGRAN]>=FIRETHRESH) hfire[yt][xp]=1;
					}
				}
			}
		}
	}
	for (int xt=1; xt<MAP_W; xt++) {
		for (int yt=0; yt<MAP_H; yt++) {
			if (iscoalright(map[yt][xt-1]) || iscoalleft(map[yt][xt])) {
				if (walkable(map[yt][xt-1]) || walkable(map[yt][xt])) {
					int yp=1+yt*FIREGRAN;
					int ypend=yp+FIREGRAN;
					if (vfire[xt][yp]==0) {
						if (hfire[yt][xt*FIREGRAN]>=FIRETHRESH || hfire[yt][1+xt*FIREGRAN]>=FIRETHRESH) vfire[xt][yp]=1;
					}
					for (; yp<ypend; yp++) {
						if (vfire[xt][yp]==0) {
							if (vfire[xt][yp-1]>=FIRETHRESH || vfire[xt][yp+1]>=FIRETHRESH) vfire[xt][yp]=1;
						}
					}
					yp--;
					if (vfire[xt][yp]==0) {
						if (hfire[yt+1][xt*FIREGRAN]>=FIRETHRESH || hfire[yt+1][1+xt*FIREGRAN]>=FIRETHRESH) vfire[xt][yp]=1;
					}
				}
			}
		}
	}
	//Move players and fire shots
	for (int i=0; i<4; i++) {
		if (chr[i].active) {
			chr[i].move(this,i,playerkeys&((1<<PLAYER_KSHIFT)-1),0,chr[i].health<=0);
			chr[i].animate(this,i);
		}
		playerkeys>>=PLAYER_KSHIFT;
	}
	//Move miners
	for (int i=4; i<NUMCHARS; i++) {
		if (chr[i].active) {
			chr[i].move(this,i,minerkeys&((1<<MINER_KSHIFT)-1),0,chr[i].health<=0);
			chr[i].animate(this,i);
		}
		minerkeys>>=MINER_KSHIFT;
	}
	//Hurt characters who are standing in the fire
	for (int i=0; i<NUMCHARS; i++) {
		if (chr[i].active) {
			int xpl=chr[i].x-128,xph=chr[i].x+chr[i].width+128;
			int ypl=chr[i].y-128,yph=chr[i].y+chr[i].height+256;//fire rises
			if (xpl<0) xpl=0; if (xph>=MAP_W<<8) xph=(MAP_W<<8)-1;
			if (ypl<0) ypl=0; if (yph>=MAP_H<<8) yph=(MAP_H<<8)-1;
			int xtl=(xpl+255)>>8,xth=xph>>8;
			int ytl=(ypl+255)>>8,yth=yph>>8;
			xpl=1+((xpl*FIREGRAN+255)>>8); xph=1+((xph*FIREGRAN)>>8);
			ypl=1+((ypl*FIREGRAN+255)>>8); yph=1+((yph*FIREGRAN)>>8);
			int total=0;
			for (int yt=ytl; yt<=yth; yt++) {
				for (int xp=xpl; xp<=xph; xp++) {
					total+=hfire[yt][xp];
				}
			}
			for (int xt=xtl; xt<=xth; xt++) {
				for (int yp=ypl; yp<=yph; yp++) {
					total+=vfire[xt][yp];
				}
			}
			chr[i].health-=total/chr[i].firedefence;
		}
	}
	//Process shots
	updateShots(myid,projx,projy);
	//Finish game?
	int alivecount=0;
	for (int i=0; i<4; i++) {
		if (chr[i].active) {
			if (chr[i].health>0) alivecount++;
			else if (i<4 && survivaltime[i]==0x7FFFFFFF) survivaltime[i]=time;
		}
	}
	if (alivecount<=(TEST_ONE_PLAYER?0:1)) finished=true;
	//Tick!
	time++;
}

void Game::updateShots(int myid, int projx, int projy) {
	Shot **sp=&shots;
	while (*sp!=NULL) {
		Shot *s=*sp;
		bool kill=false;
		//If shot is still flying, update it - provided game hasn't finished
		if (!finished && s->movetype!=MT_DIE && !s->hitme) {
			//Move horizontally - check against walls - start fire if chicken fired it
			s->x+=s->xv;
			if (!walkable(map[s->y>>8][s->x>>8])) {
				kill=true;
				if (s->id==3) {
					if (s->xv>0 && iscoalleft(map[s->y>>8][s->x>>8])) {
						if (vfire[s->x>>8][s->y*FIREGRAN>>8]==0) vfire[s->x>>8][s->y*FIREGRAN>>8]=1;
					}
					if (s->xv<0 && iscoalright(map[s->y>>8][s->x>>8])) {
						if (vfire[(s->x>>8)+1][s->y*FIREGRAN>>8]==0) vfire[(s->x>>8)+1][s->y*FIREGRAN>>8]=1;
					}
				}
			}
			//Move vertically - check against walls - start fire if chicken fired it
			s->y+=s->yv;
			if (!walkable(map[s->y>>8][s->x>>8])) {
				kill=true;
				if (s->id==3) {
					if (s->yv>0 && iscoaltop(map[s->y>>8][s->x>>8])) {
						if (hfire[s->y>>8][s->x*FIREGRAN>>8]==0) hfire[s->y>>8][s->x*FIREGRAN>>8]=1;
					}
					if (s->yv<0 && iscoalbottom(map[s->y>>8][s->x>>8])) {
						if (hfire[(s->y>>8)+1][s->x*FIREGRAN>>8]==0) hfire[(s->y>>8)+1][s->x*FIREGRAN>>8]=1;
					}
				}
			}
			//Check against characters (skip the one who fired it)
			for (int c=0; c<NUMCHARS; c++) {
				if (c!=s->id && chr[c].active) {
					if (s->x>=chr[c].x && s->x<chr[c].x+chr[c].width) {
						if (s->y>=chr[c].y && s->y<chr[c].y+chr[c].height) {
							kill=true;
							chr[c].health-=chr[s->id].shotdamage;
							if (c==myid) s->hitme=true;
						}
					}
				}
			}
		}
		//If a shot has hit me but I'm being projected, make it home in on my projected position
		if (s->hitme) {
			kill=false;
			int speed=chr[s->id].shotspeed;
			int xd=projx+chr[myid].width/2-s->x,yd=projy+chr[myid].height/2-s->y;
			if (xd>yd && xd>-yd) {s->xv=speed; s->yv=(speed*yd+xd/2)/xd; s->movetype=MT_RIGHT;}
			else if (xd<yd && xd<-yd) {s->xv=-speed; s->yv=(speed*yd-xd/2)/-xd; s->movetype=MT_LEFT;}
			else if (yd>0) {s->yv=speed; s->xv=(speed*xd+yd/2)/yd; s->movetype=MT_DOWN;}
			else if (yd<0) {s->yv=-speed; s->xv=(speed*xd-yd/2)/-yd; s->movetype=MT_UP;}
			s->x+=s->xv; s->y+=s->yv;
			if (s->x>=projx && s->x<projx+chr[myid].width) {
				if (s->y>=projy && s->y<projy+chr[myid].height) {
					s->hitme=false;
					kill=true;
				}
			}
		}
		//Put the shot into death animation if necessary
		if (kill) {s->movetype=MT_DIE; s->frame=-1;}
		//Animate - and unlink if it reaches the end of the dying animation
		s->frame++;
		if (s->frame>=chr[s->id].shotframecount[s->movetype]*chr[s->id].shotframetime[s->movetype]) {
			if (s->movetype==MT_DIE) {*sp=s->next; delete s; continue;}
			s->frame=0;
		}
		//Next shot
		sp=&s->next;
	}
}

void Game::updateEnding(int myid, int projx, int projy) {
	//Animate players and miners
	for (int i=0; i<NUMCHARS; i++) {
		if (chr[i].active) chr[i].animate(this,i);
	}
	//Animate shots
	updateShots(myid,projx,projy);
}

void Character::move(Game *g, int id, int keys, int ticksinadvance, bool dead) {
	//Check where I am
	bool onground=false;
	if (!walkable(g->map[(y+height)>>8][x>>8])) onground=true;
	if (!walkable(g->map[(y+height)>>8][(x+width-1)>>8])) onground=true;
	bool onladder=false;
	if (g->map[(y+height/2)>>8][(x+width/2)>>8]==T_LADDER) onladder=true;
	if (dead) onladder=false;
	int newmovetype=movetype;
	bool animate=false;
	//When on ground - allow player to run left/right
	if (onground) {
		//Run left - or decelerate left
		if (!dead && (keys&(K_LEFT|K_RIGHT))==K_LEFT) {
			newmovetype=MT_LEFT; animate=true;
			xv-=runaccel;
			if (xv<-runspeed) xv=-runspeed;
		}
		else if (xv<0) xv/=2;
		//Run right - or decelerate right
		if (!dead && (keys&(K_LEFT|K_RIGHT))==K_RIGHT) {
			newmovetype=MT_RIGHT; animate=true;
			xv+=runaccel;
			if (xv>runspeed) xv=runspeed;
		}
		else if (xv>0) xv/=2;
	}
	//When on ladder - allow player to climb left/right/up/down
	if (onladder) {
		//Climb left - or decelerate left - note, no animation
		if ((keys&(K_LEFT|K_RIGHT))==K_LEFT) {
			xv-=climbaccel;
			if (xv<-climbspeed) xv=-climbspeed;
		}
		else if (xv<0) xv/=2;
		//Climb right - or decelerate right - note, no animation
		if ((keys&(K_LEFT|K_RIGHT))==K_RIGHT) {
			xv+=climbaccel;
			if (xv>climbspeed) xv=climbspeed;
		}
		else if (xv>0) xv/=2;
		//Climb up - or decelerate up
		if ((keys&(K_UP|K_DOWN))==K_UP) {
			newmovetype=MT_UP; animate=true;
			yv-=climbaccel;
			if (yv<-climbspeed) yv=-climbspeed;
		}
		else if (yv<0) yv/=2;
		//Climb down - or decelerate down - note, cannot climb down if on ground
		if (!onground && (keys&(K_UP|K_DOWN))==K_DOWN) {
			newmovetype=MT_DOWN; animate=true;
			yv+=climbaccel;
			if (yv>climbspeed) yv=climbspeed;
		}
		else if (yv>0) yv/=2;
	}
	//If not on ground or ladder - fall
	if (!onground && !onladder) {
		yv+=FALLACCEL;
		if (yv>fallspeed) yv=fallspeed;
	}
	//Animate
	if (movetype<4) {
		if (newmovetype!=movetype) {movetype=newmovetype; moveframe=moveframetime[movetype]-1;}
		if (animate && !g->finished) {
			moveframe++;
			if (moveframe==moveframecount[movetype]*moveframetime[movetype]) {
				moveframe=moveframeloop[movetype]*moveframetime[movetype];
			}
		}
		else moveframe=0;
	}
	//Move horizontally
	x+=xv;
	if (!walkable(g->map[y>>8][x>>8]) || !walkable(g->map[(y+height-1)>>8][x>>8])) {
		x=(x&~255)+256;
		xv=0;
	}
	if (!walkable(g->map[y>>8][(x+width-1)>>8]) || !walkable(g->map[(y+height-1)>>8][(x+width-1)>>8])) {
		x=(x&~255)+256-width;
		xv=0;
	}
	//Move vertically
	y+=yv;
	if (!walkable(g->map[y>>8][x>>8]) || !walkable(g->map[y>>8][(x+width-1)>>8])) {
		y=(y&~255)+256;
		yv=0;
	}
	if (!walkable(g->map[(y+height-1)>>8][x>>8]) || !walkable(g->map[(y+height-1)>>8][(x+width-1)>>8])) {
		y=(y&~255)+256-height;
		yv=0;
	}
	//Shoot
	if (shootframe>=0) {
		shootframe++;
		if (shootframe>=shootframecount*shootframetime) shootframe=-1;
	}
	if (!dead && (keys&K_SHOOT) && shootdelay==0) {
		Shot *s=new Shot();
		s->id=id;
		s->x=x; s->y=y;
		s->xv=xv; s->yv=yv;
		if (movetype==MT_LEFT) {s->xv-=shotspeed; s->y+=height/2;}
		if (movetype==MT_RIGHT) {s->xv+=shotspeed; s->x+=width-1; s->y+=height/2;}
		if (movetype==MT_UP) {s->yv-=shotspeed; s->x+=width/2;}
		if (movetype==MT_DOWN) {s->yv+=shotspeed; s->x+=width/2; s->y+=height-1;}
		s->movetype=movetype;
		s->frame=-ticksinadvance;
		while (s->frame<0) s->frame+=shotframecount[movetype]*shotframetime[movetype];
		s->next=g->shots;
		g->shots=s;
		shootframe=0;
		shootdelay=shootinterval;
	}
	if (shootdelay>0) shootdelay--;
	//Set fire to ground if we're a dead miner
	if (id>=4 && health<=0) {
		if (iscoaltop(g->map[(y+height)>>8][(x+width/2)>>8])) {
			if (g->hfire[(y+height)>>8][(x+width/2)*FIREGRAN>>8]==0) g->hfire[(y+height)>>8][(x+width/2)*FIREGRAN>>8]=1;
		}
	}
}

void Character::animate(Game *g, int id) {
	if (g->finished || health<=0) {
		if (health<=0 && movetype!=MT_DIE) {movetype=MT_DIE; moveframe=-1;}
		if (health>0 && id<4 && movetype!=MT_WIN) {movetype=MT_WIN; moveframe=-1;}
		if (movetype>=4) {
			moveframe++;
			if (moveframe>=moveframecount[movetype]*moveframetime[movetype]) {
				moveframe=moveframeloop[movetype]*moveframetime[movetype];
			}
		}
		else moveframe=0;
	}
}

void Game::draw(int myid) {
	//Draw the map
	for (int y=0; y<MAP_H; y++) {
		for (int x=0; x<MAP_W; x++) {
			blit((BITMAP *)dat[TILES_000+map[y][x]].dat,scrbuf,0,0,x*TILE_W,y*TILE_H,TILE_W,TILE_H);
			if (walkable(map[y][x])) {
				if (x>0 && !walkable(map[y][x-1])) {
					BITMAP *spr=(BITMAP *)dat[iscoalright(map[y][x-1])?EDGELR_002:EDGELR_000].dat;
					draw_sprite(scrbuf,spr,x*TILE_W,y*TILE_H);
				}
				if (x<MAP_W-1 && !walkable(map[y][x+1])) {
					BITMAP *spr=(BITMAP *)dat[iscoalleft(map[y][x+1])?EDGELR_003:EDGELR_001].dat;
					draw_sprite(scrbuf,spr,x*TILE_W+TILE_W-2,y*TILE_H);
				}
				if (y>0 && !walkable(map[y-1][x])) {
					BITMAP *spr=(BITMAP *)dat[iscoalbottom(map[y-1][x])?EDGETB_002:EDGETB_000].dat;
					draw_sprite(scrbuf,spr,x*TILE_W,y*TILE_H);
				}
				if (y<MAP_H-1 && !walkable(map[y+1][x])) {
					BITMAP *spr=(BITMAP *)dat[iscoaltop(map[y+1][x])?EDGETB_003:EDGETB_001].dat;
					draw_sprite(scrbuf,spr,x*TILE_W,y*TILE_H+TILE_H-2);
				}
			}
		}
	}
	//Draw shots
	for (Shot *s=shots; s!=NULL; s=s->next) s->draw(this);
	//Draw the miners
	for (int i=4; i<NUMCHARS; i++) {
		if (chr[i].active) chr[i].draw();
	}
	//Draw the other players
	for (int i=0; i<4; i++) {
		if (i!=myid && chr[i].active) chr[i].draw();
	}
	//Draw me
	if (chr[myid].active) chr[myid].draw();
}

void Character::draw() {
	int datid;
	if (shootframe>=0 && movetype<4) datid=shootframebase[movetype]+shootframe/shootframetime;
	else datid=moveframebase[movetype]+moveframe/moveframetime[movetype];
	BITMAP *spr=(BITMAP *)dat[datid].dat;
	int xscr=(x*TILE_W>>8)+((width*TILE_W>>8)-spr->w)/2;
	int yscr=((y+height)*TILE_H>>8)-spr->h;
	draw_sprite(scrbuf,spr,xscr,yscr);
}

void Shot::draw(Game *g) {
	int datid=g->chr[id].shotframebase[movetype]+frame/g->chr[id].shotframetime[movetype];
	BITMAP *spr=(BITMAP *)dat[datid].dat;
	int xscr=(x*TILE_W>>8)-spr->w/2;
	int yscr=(y*TILE_H>>8)-spr->h/2;
	if (movetype<4) {
		int angle=0;
		if (xv!=0 || yv!=0) angle=fixatan2(yv,xv);
		rotate_sprite(scrbuf,spr,xscr,yscr,angle);
	}
	else draw_sprite(scrbuf,spr,xscr,yscr);
}
