#define ALLEGRO_USE_CONSOLE

#include <cstdlib>
#include <cstdio>
#include <allegro.h>
#include <nl.h>
#include "pyrotrash.h"

#define PRINT_KEYQUEUE_STATUS false
#define SERVER_SHOWS_GAME false

//HawkNL's macros suck - add brackets
#undef writeByte
#undef writeBlock
#undef readByte
#undef readBlock
#define writeByte(x, y, z)      (*(NLbyte *)&x[y++] = (NLbyte)(z))
#define writeBlock(x, y, z, a)  {memcpy((char *)&x[y], (char *)(z), a);y += (a);}
#define readByte(x, y, z)       (z = *(NLbyte *)&x[y++])
#define readBlock(x, y, z, a)   {memcpy((char *)(z), (char *)&x[y], a);y += (a);}

#define XMARGIN 8//128
#define YMARGIN 8//96

#define BURSTLEN 25

using namespace std;

static void usage(char *name) {
	printf("Usage: %s server <port> <maxconnections>\n",name);
	printf("- Runs a server and listens on the specified TCP port.\n");
	printf("- Note: server can host multiple games! Specify max total connections.\n");
	printf("Usage: %s <ip/host> <port> [-windowed]\n",name);
	printf("- Runs the game and connects to the server specified.\n");
	exit(EXIT_FAILURE);
}

static void server(int maxplayers);
static void client();

static int gfxmode=GFX_AUTODETECT;
static const char *host;
static int port;

int main(int argc, char *argv[]) {
	if (argc!=3 && argc!=4) usage(argv[0]);
	port=(int)strtol(argv[2],NULL,10);
	host=argv[1];
	if (stricmp(host,"server")==0) {
		if (argc!=4) usage(argv[0]);
		int maxplayers=(int)strtol(argv[3],NULL,10);
		server(maxplayers);
	} else {
		if (argc==4) {
			if (stricmp(argv[3],"-windowed")!=0 && stricmp(argv[3],"--windowed")!=0) usage(argv[0]);
			gfxmode=GFX_AUTODETECT_WINDOWED;
		}
		client();
	}
	return EXIT_SUCCESS;
}
END_OF_MAIN()

static void doNlError(const char *when) {
	if (nlGetError()==NL_SYSTEM_ERROR) {
		printf("System error %s: %s\n",when,nlGetSystemErrorStr(nlGetSystemError()));
	} else {
		printf("HawkNL error %s: %s\n",when,nlGetErrorStr(nlGetError()));
	}
	exit(EXIT_FAILURE);
}

//Packets
//-------

#define VERSION "Pyrotrash v1"

#define KEEPALIVE 0

#define C2S_CLOSING 1
#define C2S_VERSION 2
#define C2S_MY_NAME 3
#define C2S_GAME_NAME 4
#define C2S_CHOOSE_CHARACTER 5
//player number added to that, so 6,7,8 used up
#define C2S_READY 9
#define C2S_KEYSTATES 10
#define C2S_CHAT 11

#define S2C_SERVER_FULL 1
#define S2C_WRONG_VERSION 2
#define S2C_GAME_SLOTS 3
#define S2C_PLAYER_READY 4
//player number added to that, so 5,6,7 used up
#define S2C_PLAYER_WAITING 8
//player number added to that, so 9,10,11 used up
#define S2C_PLAYER_LEFT 12
//player number added to that, so 13,14,15 used up
#define S2C_READY 16
#define S2C_STEADY 17
#define S2C_GO 18
#define S2C_CANCEL_START 19
#define S2C_MYID 20
//player number added to that, so 21,22,23 used up
#define S2C_START_GAME 24
//Horrible hack - S2C_START_GAME given size of two bytes, but other packets have one
#define S2C_KEYSTATES 25
#define S2C_CHAT 26

//Server
//------

#ifdef ALLEGRO_WINDOWS
#	include <winalleg.h>
#	define YIELD(n) Sleep(n)
#elif defined ALLEGRO_UNIX
#	include <sys/time.h>
	static void YIELD(int n) {
		struct timeval tv;
		tv.tv_sec=n/1000;
		tv.tv_usec=(n%1000)*1000;
		select(0,NULL,NULL,NULL,&tv);
	}
#else
#	error The server will use 100% CPU unless you fix this! Comment this error line out if you don`t care.
#	define YIELD(n) nlThreadYield()
#endif

struct servergame;

struct serverplayer {
	serverplayer *next;
	NLsocket socket;
	int inTimeout,outTimeout;
	unsigned char in[1025];//was 257 - needs to be bigger for S2C_START_GAME
	int inlen;
	unsigned char out[1024];
	int outpos;
	char name[256];
	bool notifyactive;
	servergame *notifygame;
	int notifyknown;
	servergame *game;
	int id;//0-3 depending on who you are in the game you're in
	bool readin(int callsPerSec, int longptype);
	void flushout(int callsPerSec);
};

bool serverplayer::readin(int callsPerSec, int longptype) {
	inTimeout++;
	if (inTimeout>=callsPerSec*30) {nlClose(socket); socket=NL_INVALID; return false;}
	if (inlen<2) {
		int read=nlRead(socket,in+inlen,2-inlen);
		if (read==NL_INVALID) {nlClose(socket); socket=NL_INVALID; return false;}
		inlen+=read;
		//if (inlen>0) printf("Player %s: read %d of 2 (header)\n",name,inlen);
		if (inlen<2) return false;
	}
	if (in[0]==longptype && inlen<3) {
		int read=nlRead(socket,in+inlen,3-inlen);
		if (read==NL_INVALID) {nlClose(socket); socket=NL_INVALID; return false;}
		inlen+=read;
		//if (inlen>0) printf("Player %s: read %d of 3 (header)\n",name,inlen);
		if (inlen<3) return false;
	}
	int len;
	if (in[0]==longptype) len=(((in[1]<<8)+in[2])&1023)+3;
	else len=in[1]+2;
	if (inlen<len) {
		int read=nlRead(socket,in+inlen,len-inlen);
		if (read==NL_INVALID) {nlClose(socket); socket=NL_INVALID; return false;}
		inlen+=read;
		//printf("Player %s: read %d of %d (header+body)\n",name,inlen,len);
		if (inlen<len) return false;
	}
	inTimeout=0;
	return true;
}

void serverplayer::flushout(int callsPerSec) {
	if (outpos==0) {
		outTimeout++;
		if (outTimeout<callsPerSec*10) return;
		writeByte(out,outpos,KEEPALIVE);
		writeByte(out,outpos,0);
	}
	outTimeout=0;
	int written=0;
	while (true) {
		int rv=nlWrite(socket,out+written,outpos-written);
		if (rv==NL_INVALID) {
			if (nlGetError()!=NL_CON_PENDING) socket=NL_INVALID;
			return;
		}
		written+=rv;
		//printf("Player %s: written %d of %d\n",name,written,outpos);
		if (written>=outpos) {outpos=0; return;}
		YIELD(1);
	}
}

struct servergame {
	servergame *next;
	char name[256];
	int notifycount;
	int playercount;
	serverplayer *player[4];
	bool playerready[4];
	int countdown;
	Game *game;
	unsigned char keyqueue[4][512];
	int keyfirst[4],keylast[4];
};

static servergame *gamelist;

static servergame *createNewGame(const char *gamename);
static void addPlayerToGame(serverplayer *p, servergame *g, int id);
static void clearNotify(serverplayer *p);
static void removePlayerFromGame(serverplayer *p);
static void startGame(servergame *g);
static void sendchat(servergame *g, const char *chat);

static void require(const char *error, int rv);
static void require(const char *error, void *rv);

static void server(int maxplayers) {
	if (SERVER_SHOWS_GAME) {
		allegro_init();
		set_config_file("pyrotrash.ini");
		install_keyboard();
		require("load_datafile",dat=load_datafile("media/pyrotrash.dat"));
		require("set_gfx_mode",set_gfx_mode(GFX_AUTODETECT_WINDOWED,640,480,0,0));
		if (set_display_switch_mode(SWITCH_BACKGROUND)!=0) {
			set_display_switch_mode(SWITCH_BACKAMNESIA);
		}
		set_palette((RGB *)dat[THE_PALETTE].dat);
		scrbuf=create_bitmap(640,480);
	}
	printf("Pyrotrash server!\n");
	srand(time(0));
	if (!nlInit()) doNlError("initialising");
	if (!nlSelectNetwork(NL_IP)) doNlError("selecting IP");
	if (!nlEnable(NL_TCP_NO_DELAY)) doNlError("setting TCP no delay mode");
	NLsocket listener=nlOpen((NLushort)port,NL_RELIABLE);
	if (listener==NL_INVALID) doNlError("opening socket");
	if (!nlListen(listener)) doNlError("listening on socket");
	NLaddress addr;
	nlGetLocalAddr(listener,&addr);
	char str[NL_MAX_STRING_LENGTH];
	printf("Listening on %s, max %d connections!\n",nlAddrToString(&addr,str),maxplayers);
	serverplayer *playerlist=NULL;
	int playercount=0;
	gamelist=NULL;
	NLtime ts;
	if (!nlTime(&ts)) doNlError("getting time");
	int time=(int)(ts.seconds*1000+ts.mseconds);
	while (true) {
		//Accept new connections
		for (int i=0; i<10; i++) {
			NLsocket socket=nlAcceptConnection(listener);
			if (socket==NL_INVALID) {
				if (nlGetError()!=NL_NO_PENDING) doNlError("accepting connection");
				break;
			}
			serverplayer *p=new serverplayer();
			p->socket=socket;
			p->inTimeout=p->outTimeout=0;
			p->inlen=p->outpos=0;
			p->name[0]=0;
			p->notifyactive=false;
			p->notifygame=NULL;
			p->game=NULL;
			if (playercount>=maxplayers) {
				writeByte(p->out,p->outpos,S2C_SERVER_FULL);
				writeByte(p->out,p->outpos,0);
				p->flushout(100/BURSTLEN);
				nlClose(socket);
				delete p;
				continue;
			}
			p->next=playerlist; playerlist=p;
			playercount++;
		}
		//Process connections: timeout and packet decoding
		//Also process notifies in this loop
		serverplayer **pp=&playerlist;
		while (true) {
			serverplayer *p=*pp;
			if (p==NULL) break;
			//Get incoming packets
			while (p->readin(100/BURSTLEN,-1)) {
				int inpos=0;
				int type,len;
				readByte(p->in,inpos,type); type&=255;
				readByte(p->in,inpos,len); len&=255;
				if (type==KEEPALIVE) {
				}
				else if (type==C2S_VERSION) {
					char version[256];
					readBlock(p->in,inpos,version,len);
					version[len>0?len-1:0]=0;
					if (strcmp(version,VERSION)!=0) {
						writeByte(p->out,p->outpos,S2C_WRONG_VERSION);
						writeByte(p->out,p->outpos,strlen(VERSION)+1);
						writeBlock(p->out,p->outpos,VERSION,strlen(VERSION)+1);
						p->flushout(100/BURSTLEN);
						nlClose(p->socket); p->socket=NL_INVALID;
						break;
					}
				}
				else if (type==C2S_MY_NAME) {
					if (len>150) len=150;//for security safety - allows chat with name + about 100 extra chars
					readBlock(p->in,inpos,p->name,len);
					p->name[len>0?len-1:0]=0;
				}
				else if (type==C2S_GAME_NAME) {
					char gamename[256];
					readBlock(p->in,inpos,gamename,len);
					gamename[len>0?len-1:0]=0;
					if (p->game==NULL) {
						clearNotify(p);
						p->notifyactive=true;
						if (gamename[0]!=0) {
							servergame *g=NULL;
							for (servergame *g2=gamelist; g2!=NULL; g2=g2->next) {
								if (stricmp(gamename,g2->name)==0) {g=g2; break;}
							}
							if (g==NULL) g=createNewGame(gamename);
							p->notifygame=g; g->notifycount++;
						}
						p->notifyknown=-1;
					}
				}
				else if (type>=C2S_CHOOSE_CHARACTER && type<C2S_CHOOSE_CHARACTER+4) {
					int id=type-C2S_CHOOSE_CHARACTER;
					addPlayerToGame(p,p->notifygame,id);
				}
				else if (type==C2S_KEYSTATES) {
					if (p->game!=NULL && p->game->game!=NULL) {
						int &keyfirst=p->game->keyfirst[p->id];
						int &keylast=p->game->keylast[p->id];
						if (keylast-keyfirst+len>511) len=511-keylast+keyfirst;
						for (int i=0; i<len; i++) {
							readByte(p->in,inpos,p->game->keyqueue[p->id][keylast&511]);
							keylast++;
						}
						if (PRINT_KEYQUEUE_STATUS) printf("%s: keyqueue at %d/511\n",p->name,keylast-keyfirst);
					}
				}
				else if (type==C2S_CHAT) {
					if (p->game!=NULL) {
						char chat[256];
						int pos=sprintf(chat,"<%s> ",p->name);
						if (pos+len>255) len=255-pos;
						readBlock(p->in,inpos,chat+pos,len);
						len+=pos;
						chat[len-1]=0;
						sendchat(p->game,chat);
					}
				}
				else if (type==C2S_READY) {
					if (p->game!=NULL && p->game->game==NULL) {
						p->game->playerready[p->id]=true;
						//Tell all other players that we're ready to play again
						for (int j=0; j<4; j++) {
							serverplayer *destp=p->game->player[j];
							if (destp!=NULL && destp!=p) {
								writeByte(destp->out,destp->outpos,S2C_PLAYER_READY+p->id);
								writeByte(destp->out,destp->outpos,strlen(p->name)+1);
								writeBlock(destp->out,destp->outpos,p->name,strlen(p->name)+1);
								//printf("Told %s player %d is %s\n",destp->name,i,p->name);
							}
						}
					}
				}
				else { // C2S_CLOSING, or unrecognised packet
					nlClose(p->socket); p->socket=NL_INVALID;
					break;
				}
				p->inlen=0;
			}
			//Send updates for any game we're watching
			if (p->socket!=NL_INVALID && p->notifyactive) {
				int newnotify=0;
				servergame *notifygame=p->notifygame;
				//If no notifygame, find unnamed game that isn't full or active
				if (notifygame==NULL) {
					//Find the LAST matching game in the list (i.e. the earliest added, the oldest)
					for (servergame *g2=gamelist; g2!=NULL; g2=g2->next) {
						if (g2->playercount<4 && g2->game==NULL && g2->name[0]==0) notifygame=g2;
					}
				}
				if (notifygame!=NULL) {
					if (notifygame->game!=NULL) newnotify=15;
					else {
						for (int i=0; i<4; i++) {
							if (notifygame->player[i]!=NULL) newnotify+=1<<i;
						}
					}
				}
				if (newnotify!=p->notifyknown) {
					writeByte(p->out,p->outpos,S2C_GAME_SLOTS);
					writeByte(p->out,p->outpos,1);
					writeByte(p->out,p->outpos,newnotify);
					p->notifyknown=newnotify;
				}
			}
			//Delete this connection if dead, and move on to the next connection
			if (p->socket==NL_INVALID) {*pp=p->next; removePlayerFromGame(p); delete p; playercount--; continue;}
			pp=&p->next;
		}
		//Process games
		servergame **gp=&gamelist;
		while (true) {
			servergame *g=*gp;
			if (g==NULL) break;
			//Check for dead game room
			if (g->notifycount==0 && g->playercount==0) {
				*gp=g->next;
				if (g->game!=NULL) delete g->game;
				delete g;
				continue;
			}
			//Check for game that can start
			if (g->game==NULL) {
				bool ready;
				if (g->playercount>=(TEST_ONE_PLAYER?1:2)) {
					ready=true;
					for (int i=0; i<4; i++) {
						if (g->player[i]!=NULL && !g->playerready[i]) ready=false;
					}
				} else ready=false;
				if (ready) {
		 			if (g->countdown==0) g->countdown=3*(100/BURSTLEN); else g->countdown--;
					if (g->countdown!=0 && (g->countdown&3)==0) {
						int type;
						if (g->countdown==3*(100/BURSTLEN)) type=S2C_READY;
						else if (g->countdown==2*(100/BURSTLEN)) type=S2C_STEADY;
						else type=S2C_GO;
						for (int i=0; i<4; i++) {
							serverplayer *p=g->player[i];
							if (p!=NULL) {
								writeByte(p->out,p->outpos,type);
								writeByte(p->out,p->outpos,0);
							}
						}
					}
					if (g->countdown==0) startGame(g);
				} else {
					if (g->countdown!=0) {
						g->countdown=0;
						for (int i=0; i<4; i++) {
							serverplayer *p=g->player[i];
							if (p!=NULL) {
								writeByte(p->out,p->outpos,S2C_CANCEL_START);
								writeByte(p->out,p->outpos,0);
							}
						}
					}
				}
			}
			//Process BURSTLEN frames of existing game
			if (g->game!=NULL) {
				int packetstart[4];
				for (int i=0; i<4; i++) packetstart[i]=0;
				for (int framecounter=0; framecounter<BURSTLEN; framecounter++) {
					//If we've done 20 frames, check if we can stop
					/* - commented out since the server should be the authority on game time!
					if (framecounter>=20) {
						bool stop=false;
						for (int i=0; i<4; i++) {
							//Stop if there are players who are present and haven't submitted any more key states
							if (g->player[i]!=NULL && g->keylast[i]==g->keyfirst[i]) stop=true;
						}
						if (stop) break;
					}
					*/
					//Read keys from each player (where available)
					int playerkeys=0;
					for (int i=0; i<4; i++) {
						int &keyfirst=g->keyfirst[i],&keylast=g->keylast[i];
						int keys;
						if (keyfirst<keylast) keys=g->keyqueue[i][keyfirst++&511];
						else keys=K_SERVER_ENFORCED;
						playerkeys+=keys<<(i*PLAYER_KSHIFT);
					}
					//Update game
					int minerkeys=g->game->getMinerKeys();
					g->game->update(playerkeys,minerkeys,-1,0,0);
					//Send these keys out to all players!
					for (int i=0; i<4; i++) {
						serverplayer *p=g->player[i];
						if (p!=NULL) {
							//Finish old KEYSTATES packet if it's getting too big
							if (packetstart[i]!=0 && p->outpos-packetstart[i]>=240) {
								p->out[packetstart[i]-1]=p->outpos-packetstart[i];
								packetstart[i]=0;
							}
							//Start new KEYSTATES packet if necessary
							if (packetstart[i]==0) {
								writeByte(p->out,p->outpos,S2C_KEYSTATES);
								p->outpos++;
								packetstart[i]=p->outpos;
							}
							writeByte(p->out,p->outpos,playerkeys);
							writeByte(p->out,p->outpos,playerkeys>>8);
							writeByte(p->out,p->outpos,playerkeys>>16);
							if (NUMCHARS>4) writeByte(p->out,p->outpos,minerkeys);
							if (NUMCHARS>6) writeByte(p->out,p->outpos,minerkeys>>8);
							if (NUMCHARS>8) writeByte(p->out,p->outpos,minerkeys>>16);
							if (NUMCHARS>10) writeByte(p->out,p->outpos,minerkeys>>24);
							/*TODO: hopefully we won't need anything like this
							//If the player has got new items, send them to that player (stop KEYSTATES packet if necessary)
							if (g->itemcount[i]>g->knownitemcount[i]) {
								if (packetstart[i]!=0) {p->out[packetstart[i]-1]=p->outpos-packetstart[i]; packetstart[i]=0;}
								writeByte(p->out,p->outpos,S2C_ITEMS);
								writeByte(p->out,p->outpos,g->itemcount[i]-g->knownitemcount[i]);
								for (int j=g->knownitemcount[i]; j<g->itemcount[i]; j++) {
									writeByte(p->out,p->outpos,g->items[i][j]);
								}
								g->knownitemcount[i]=g->itemcount[i];
							}
							*/
						}
					}
					if (g->game->finished) break;
				}
				//Finish KEYSTATES packets if necessary
				for (int i=0; i<4; i++) {
					if (packetstart[i]!=0) {
						serverplayer *p=g->player[i];
						p->out[packetstart[i]-1]=p->outpos-packetstart[i];
						packetstart[i]=0;
					}
				}
				if (SERVER_SHOWS_GAME) {
					clear(scrbuf);
					g->game->draw(-1);
					blit(scrbuf,screen,0,0,0,0,640,480);
				}
				if (PRINT_KEYQUEUE_STATUS) {
					for (int i=0; i<4; i++) {
						serverplayer *p=g->player[i];
						int &keyfirst=g->keyfirst[i],&keylast=g->keylast[i];
						if (p!=NULL) printf("%s: keyqueue at %d/511\n",p->name,keylast-keyfirst);
					}
				}
				//Game finished?
				if (g->game->finished) {
					//TODO: we could finish the processing here and do ratings and whatnot, but for a TINS? Um, NO
					delete g->game;
					g->game=NULL;
					for (int i=0; i<4; i++) g->playerready[i]=false;
				}
			}
			gp=&g->next;
		}
		//Process connections: flushout and keepalive
		pp=&playerlist;
		while (true) {
			serverplayer *p=*pp;
			if (p==NULL) break;
			//Flush outgoing packets
			p->flushout(100/BURSTLEN);
			if (p->socket==NL_INVALID) {*pp=p->next; removePlayerFromGame(p); delete p; playercount--; continue;}
			pp=&p->next;
		}
		//Tick the server in line with BURSTLEN
		time+=BURSTLEN*10;
		while (true) {
			if (!nlTime(&ts)) doNlError("getting time");
			int timenow=(int)(ts.seconds*1000+ts.mseconds);
			if (time-timenow<=0) break;
			YIELD(time-timenow);
		}
	}
	if (!nlClose(listener)) doNlError("closing socket");
	nlShutdown();
}

static servergame *createNewGame(const char *gamename) {
	servergame *g=new servergame();
	strcpy(g->name,gamename);
	g->notifycount=0;
	g->playercount=0;
	for (int i=0; i<4; i++) g->player[i]=NULL;
	g->next=gamelist; gamelist=g;
	return g;
}

static void addPlayerToGame(serverplayer *p, servergame *g, int id) {
	if (p->game!=NULL) return;
	//If no game passed in, then find appropriate random game
	if (g==NULL) {
		//Find the LAST matching game in the list (i.e. the earliest added, the oldest)
		for (servergame *g2=gamelist; g2!=NULL; g2=g2->next) {
			if (g2->playercount<4 && g2->game==NULL && g2->name[0]==0) g=g2;
		}
		if (g==NULL) g=createNewGame("");
	}
	//Check that the ID the player wants is free, and the game is not in progress
	if (g->player[id]!=NULL || g->game!=NULL) return;
	//This will be a successful add - clear notify now
	clearNotify(p);
	//Tell us my ID
	writeByte(p->out,p->outpos,S2C_MYID+id);
	writeByte(p->out,p->outpos,0);
	//Tell us who's in this game already
	for (int i=0; i<4; i++) {
		if (g->player[i]!=NULL) {
			const char *pname=g->player[i]->name;
			writeByte(p->out,p->outpos,(g->playerready[i]?S2C_PLAYER_READY:S2C_PLAYER_WAITING)+i);
			writeByte(p->out,p->outpos,strlen(pname)+1);
			writeBlock(p->out,p->outpos,pname,strlen(pname)+1);
			//printf("Told %s player %d is %s\n",p->name,i,pname);
		}
	}
	//Add us to this game
	g->player[id]=p; g->playerready[id]=false; g->playercount++;
	p->game=g; p->id=id;
	//Tell all players including us that we're in this game
	for (int j=0; j<4; j++) {
		serverplayer *destp=g->player[j];
		if (destp!=NULL) {
			writeByte(destp->out,destp->outpos,S2C_PLAYER_WAITING+id);
			writeByte(destp->out,destp->outpos,strlen(p->name)+1);
			writeBlock(destp->out,destp->outpos,p->name,strlen(p->name)+1);
			//printf("Told %s player %d is %s\n",destp->name,i,p->name);
		}
	}
}

static void clearNotify(serverplayer *p) {
	p->notifyactive=false;
	if (p->notifygame!=NULL) {
		p->notifygame->notifycount--;
		p->notifygame=NULL;
	}
}

//Note: this is only called when the player is being unlinked, so stuff in 'p' is left unresolved
static void removePlayerFromGame(serverplayer *p) {
	clearNotify(p);
	if (p->game==NULL) return;
	servergame *g=p->game;
	g->player[p->id]=NULL; g->playercount--;
	//Tell other players
	for (int i=0; i<4; i++) {
		serverplayer *otherp=g->player[i];
		if (otherp!=NULL) {
			writeByte(otherp->out,otherp->outpos,S2C_PLAYER_LEFT+p->id);
			writeByte(otherp->out,otherp->outpos,0);
		}
	}
}

static void startGame(servergame *g) {
	for (int i=0; i<4; i++) g->keyfirst[i]=g->keylast[i]=0;
	//Build the game
	g->game=new Game();
	g->game->generateMap();
	for (int i=0; i<4; i++) {
		if (g->player[i]!=NULL) g->game->placePlayer(i);
	}
	g->game->placeMiners();
	//Send the initial state out
	for (int i=0; i<4; i++) {
		serverplayer *p=g->player[i];
		if (p!=NULL) {
			writeByte(p->out,p->outpos,S2C_START_GAME);
			int start=p->outpos;
			p->outpos+=2;
			for (int y=0; y<MAP_H; y++) writeBlock(p->out,p->outpos,g->game->map[y],MAP_W);
			for (int c=0; c<NUMCHARS; c++) {
				if (g->game->chr[c].active) {
					writeByte(p->out,p->outpos,1);
					writeShort(p->out,p->outpos,g->game->chr[c].x);
					writeShort(p->out,p->outpos,g->game->chr[c].y);
					writeByte(p->out,p->outpos,g->game->chr[c].movetype);
				}
				else writeByte(p->out,p->outpos,0);
			}
			int size=p->outpos-start-2;
			writeShort(p->out,start,size);
		}
	}
}

static void sendchat(servergame *game, const char *chat) {
	int len=strlen(chat)+1;
	for (int i=0; i<4; i++) {
		serverplayer *p=game->player[i];
		if (p!=NULL) {
			writeByte(p->out,p->outpos,S2C_CHAT);
			writeByte(p->out,p->outpos,len);
			writeBlock(p->out,p->outpos,chat,len);
		}
	}
}

//Client
//------

static void require(const char *error, int rv) {
	if (rv) {
		set_gfx_mode(GFX_TEXT,0,0,0,0);
		allegro_message("%s failed\n", error);
		exit(1);
	}
}

static void require(const char *error, void *rv) {
	if (!rv) {
		set_gfx_mode(GFX_TEXT,0,0,0,0);
		allegro_message("%s failed\n", error);
		exit(1);
	}
}

static volatile int truetick=0;
static int logictick=0;
static int maxlogictick=10;

DATAFILE *dat;

BITMAP *scrbuf;

static void timerf() {truetick++;}
END_OF_STATIC_FUNCTION(timerf)

static void clientInit();
static void clientDraw();
static void clientUpdate();

static void client() {
	if (!nlInit()) doNlError("initialising");
	if (!nlSelectNetwork(NL_IP)) doNlError("selecting IP");
	if (!nlEnable(NL_TCP_NO_DELAY)) doNlError("setting TCP no delay mode");
	LOCK_VARIABLE(truetick);
	LOCK_FUNCTION(timerf);
	srand(time(0));
	allegro_init();
	set_config_file("pyrotrash.ini");
	install_keyboard();
	install_timer();
	//reserve_voices(64,-1);
	//set_volume_per_voice(0);
	//install_sound(DIGI_AUTODETECT,MIDI_NONE,0);
	//dumb_register_dat_it_quick(DUMB_DAT_IT);
	require("load_datafile",dat=load_datafile("media/pyrotrash.dat"));
	require("set_gfx_mode",set_gfx_mode(gfxmode,640,480,0,0));
	if (set_display_switch_mode(SWITCH_BACKGROUND)!=0) {
		set_display_switch_mode(SWITCH_BACKAMNESIA);
	}
	set_palette((RGB *)dat[THE_PALETTE].dat);
	scrbuf=create_bitmap(640,480);
	clientInit();
	install_int(timerf,10);
	while (true) {
		clientDraw();
		blit(scrbuf,screen,0,0,0,0,640,480);
		while (truetick<=logictick) rest(1);
		while (truetick>logictick && logictick<maxlogictick) {clientUpdate(); logictick++;}
		if (truetick>maxlogictick) truetick=maxlogictick;
		maxlogictick=logictick+10;
	}
}

#define GS_ENTERING_NAME 0
#define GS_MAIN_MENU 1
#define GS_INSTRUCTIONS 2
#define GS_ENTERING_GAME_NAME 3
#define GS_CONNECT 4
#define GS_CONNECT_ERROR 5
#define GS_CHOOSING_CHARACTER 6
#define GS_SHOW_GAME_START 7
#define GS_PLAYER_LIST 7
#define GS_GAME 8

static int gamestate;
static char myname[256];
static int mynamelen;
static int menuopt;
static char gamename[256];
static int gamenamelen;

static int connectstep;
static NLaddress address;
static serverplayer con;
static char connecterror[NL_MAX_STRING_LENGTH];

static int gameslots;

static char playername[4][256];
static char playername2[4][256];//Persistent during game - even when players leave
static bool playerready[4];
static int readySteadyMessage;

static Game *thegame;
static Fire *thefire;
static int myid;
static int myprojx,myprojy;

static unsigned char outkeyqueue[512];
static int outkeyfirst,outkeylast;
static int outkeylastsent;

static int inplayerkeyqueue[512];
static int inminerkeyqueue[512];
static int inkeyfirst,inkeylast;
static int inlenhistory[256];
static int inlenptr;
static int inminlen;

static void exitgame();
static void clientConnectError();
static void useinkey();

static void clientInit() {
	con.socket=NL_INVALID;
	gamestate=GS_ENTERING_NAME;
	mynamelen=0; myname[0]=0;
	Fire::init();
}

static void initClientGame() {
	outkeyfirst=outkeylast=outkeylastsent=0;
	inkeyfirst=inkeylast=0;
	inlenptr=0;
	inminlen=0;
	for (int i=0; i<256; i++) inlenhistory[i]=0;
	for (int i=0; i<4; i++) strcpy(playername2[i],playername[i]);
}

static Character tmpchr;

static const int leftedge=640-100-4-100;
static const int meterx=leftedge+100+4;
static const int meterw=640-meterx;
static const int maxchatwidth=leftedge-4;

static const int chatrows=5;
static char thechat[chatrows][256];
static int myminchatlen,newchatlen;
static char newchat[256];

static const char *charname[]={
	"Crisp packet",
	"Milk carton",
	"Energy drink",
	"Burnt chicken"
};
static const char *chardesc1[]={
	"Speed: moderate   -   Attack: crisp storm",
	"Speed: fast   -   Attack: sour milk.",
	"Speed: very fast   -   Attack: caffeine beam   -   Special: retards fire",
	"Speed: slow   -   Attack: flaming stuffing   -   Special: retards fire"
};
static const char *chardesc2[]={
	"A good all-round choice.",
	"Makes up for a slightly weaker weapon with its agility.",
	"Zip around the mine like nobody's business.",
	"For the arsonist in you."
};

static void clientDraw() {
	clear(scrbuf);
	if ((gamestate<GS_SHOW_GAME_START || thegame==NULL) && gamestate!=GS_INSTRUCTIONS) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"~ Pyrotrash ~",320,100-60,-1,-1);
		if (gamestate!=GS_PLAYER_LIST) {
			textout_centre_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,"Copyright (C) 2007 Ben 'Bruce \"entheh\" Perry' Davis - TINS 2007 - amarillion.bafsoft.net/tins07",320,500-60,-1,-1);
		}
	}
	if (gamestate==GS_ENTERING_NAME) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Enter your name",320,200-60,-1,-1);
		int x=320-text_length((FONT *)dat[FONT_LARGE].dat,myname)/2;
		if (logictick&16) {myname[mynamelen]='_'; myname[mynamelen+1]=0;}
		textout_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,myname,x,240-60,-1,-1);
		myname[mynamelen]=0;
	}
	if (gamestate==GS_MAIN_MENU) {
		textprintf_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,320,200-60,-1,-1,"Welcome %s",myname);
		rectfill(scrbuf,0,235+menuopt*40-60,639,274+menuopt*40-60,2);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Instructions",320,240-60,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Join random game",320,280-60,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Join named game",320,320-60,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Change name",320,360-60,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Exit",320,400-60,-1,-1);
	}
	if (gamestate==GS_INSTRUCTIONS) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"~ Pyrotrash ~",320,20,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"\"WEST VIRGINIAN MINE FIRE KILLS TWO\"",320,80,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"\"FIRE UNDERGROUND: Reckless disregard for safety",320,120,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"results in record fine\"",320,160,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"\"$1.5 MILLION PENALTY FOR FATAL MINE FIRE\"",320,200,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"We are about to show you what really happened.",320,260,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Arrow keys to move. Space to fire.",320,320,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"I wrote this whole game before testing any of it.",320,380,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"I suspect I am officially insane.",320,420,-1,-1);
	}
	if (gamestate==GS_ENTERING_GAME_NAME) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Enter game name",320,200-60,-1,-1);
		int x=320-text_length((FONT *)dat[FONT_LARGE].dat,gamename)/2;
		if (logictick&16) {gamename[gamenamelen]='_'; gamename[gamenamelen+1]=0;}
		textout_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,gamename,x,240-60,-1,-1);
		gamename[gamenamelen]=0;
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"If you and 1-3 friends enter the same game name,",320,360-60,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"you can play each other!",320,400-60,-1,-1);
	}
	if (gamestate==GS_CONNECT) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Connecting to server ...",320,300-60,-1,-1);
	}
	if (gamestate==GS_CONNECT_ERROR) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Error connecting to server",320,300-60,-1,-1);
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,connecterror,320,340-60,-1,-1);
	}
	if (gamestate==GS_CHOOSING_CHARACTER) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,gamename[0]==0?"Unnamed game":gamename,320,200-60,-1,-1);
		if (gameslots==15) {
			textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Sorry, game full or already in progress",320,280-60,-1,-1);
		}
		else {
			textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Choose character",320,280-80,-1,-1);
			for (int i=0; i<4; i++) {
				int col=-1;
				if (gameslots&(1<<i)) col=3;
				textprintf_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,100,320-80+i*40,col,-1,"%d. %s",i+1,charname[i]);
				textout_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,chardesc1[i],300,323-80+i*40,-1,-1);
				textout_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,chardesc2[i],300,337-80+i*40,-1,-1);
			}
		}
	}
	if (gamestate==GS_PLAYER_LIST) {
		textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,gamename[0]==0?"Unnamed game":gamename,320,200-80,-1,-1);
		textout_right_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playername[0],320,280-100,-1,-1);
		textout_right_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playername[1],320,320-100,-1,-1);
		textout_right_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playername[2],320,360-100,-1,-1);
		textout_right_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playername[3],320,400-100,-1,-1);
		int count=0;
		if (playername[0][0]!=0) {count++; textout_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playerready[0]?": ready":myid==0?": press Enter to start":": waiting",320,280-100,-1,-1);}
		if (playername[1][0]!=0) {count++; textout_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playerready[1]?": ready":myid==1?": press Enter to start":": waiting",320,320-100,-1,-1);}
		if (playername[2][0]!=0) {count++; textout_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playerready[2]?": ready":myid==2?": press Enter to start":": waiting",320,360-100,-1,-1);}
		if (playername[3][0]!=0) {count++; textout_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playerready[3]?": ready":myid==3?": press Enter to start":": waiting",320,400-100,-1,-1);}
		if (readySteadyMessage==S2C_READY) textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"READY",320,480-120,-1,-1);
		else if (readySteadyMessage==S2C_STEADY) textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"STEADY",320,480-120,-1,-1);
		else if (readySteadyMessage==S2C_GO) textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"GO!!",320,480-120,-1,-1);
		else if (playerready[myid] && count==1) textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Waiting for other players",320,480-120,-1,-1);
	}
	if (gamestate>=GS_SHOW_GAME_START && thegame!=NULL) {
		//Save some state
		memcpy(&tmpchr,&thegame->chr[myid],sizeof(tmpchr));
		Shot *firstshot=thegame->shots;
		//Do buffered moves prospectively - some shots may be created
		for (int i=outkeyfirst; i<outkeylast; i++) thegame->chr[myid].move(thegame,myid,outkeyqueue[i&511],i-outkeyfirst,false);
		//Update my projected position - used for shots when they hit me
		//- purely visual, so it doesn't matter that this happens in the draw function
		myprojx=thegame->chr[myid].x+thegame->chr[myid].width/2;
		myprojy=thegame->chr[myid].y+thegame->chr[myid].height/2;
		//Draw! - and pass the ghost position in!
		scrbuf->cb=480-66;
		thegame->draw(myid);
		thefire->draw();
		scrbuf->cb=480;
		//Restore state
		memcpy(&thegame->chr[myid],&tmpchr,sizeof(tmpchr));
		while (thegame->shots!=firstshot) {
			Shot *next=thegame->shots->next;
			delete thegame->shots;
			thegame->shots=next;
		}
		//Show status at the bottom of the screen
		rectfill(scrbuf,0,480-66,639,479,0);
		//Show player names and health meters
		for (int i=0; i<4; i++) {
			if (playername2[i][0]!=0) {
				textout_right_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,playername2[i],leftedge+100,480-66+11*i,playername[i][0]==0?4:-1,-1);
				int level=(meterw*thegame->chr[i].health+MAXHEALTH-1)/MAXHEALTH;
				if (level>0) rectfill(scrbuf,meterx,480-66+11*i,meterx+level-1,480-66+11*i+10,9+i);
				textout_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,charname[i],meterx+2,480-66+11*i,-1,-1);
				if (thegame->chr[i].health<=0) {
					int seconds=thegame->survivaltime[i]/100;
					textprintf_right_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,640,480-66+11*i,-1,-1,"%d:%02d",seconds/60,seconds%60);
				}
			}
		}
		//Show time elapsed
		int seconds=thegame->time/100;
		textprintf_right_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,640,480-26,-1,-1,"%d:%02d",seconds/60,seconds%60);
		//Game finished?
		if (thegame->finished) {
			//Sort survival times into order
			int order[4];
			int numplayers=0;
			for (int i=0; i<4; i++) {
				if (playername2[i][0]!=0) order[numplayers++]=i;
			}
			int *time=thegame->survivaltime;
			for (int i=numplayers-1; i>0; i--) {
				for (int j=0; j<i; j++) {
					if (time[order[j]]<time[order[j+1]]) {int tmp=order[j]; order[j]=order[j+1]; order[j+1]=tmp;}
				}
			}
			if (thegame->chr[order[0]].health>0) {
				textprintf_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,320,60-60,-1,-1,"Winner: %s",playername2[order[0]]);
			} else {
				textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Draw",320,60-60,-1,-1);
			}
			textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Survival times:",320,120-60,-1,-1);
			for (int i=0; i<numplayers; i++) {
				textout_right_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,playername2[order[i]],320,180+i*40-60,-1,-1);
				if (thegame->chr[order[i]].health>0) textout_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,": survivor",320,180+i*40-60,-1,-1);
				else {
					int seconds=time[order[i]]/100;
					textprintf_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,320,180+i*40-60,-1,-1,": %d:%02d",seconds/60,seconds%60);
				}
			}
			textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Enter to play again",320,400-60,-1,-1);
			textout_centre_ex(scrbuf,(FONT *)dat[FONT_LARGE].dat,"Esc to leave",320,440-60,-1,-1);
		}
	}
	if (gamestate>=GS_SHOW_GAME_START) {
		//Show chat
		for (int i=0; i<chatrows; i++) textout_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,thechat[i],0,480-66+11*i,-1,-1);
		if (logictick&16) {newchat[newchatlen]='_'; newchat[newchatlen+1]=0;}
		textout_ex(scrbuf,(FONT *)dat[FONT_SMALL].dat,newchat,0,480-11,-1,-1);
		newchat[newchatlen]=0;
	}
}

static void clientUpdate() {
	if (gamestate==GS_ENTERING_NAME) {
		while (keypressed()) {
			int k=readkey();
			if (k>>8==KEY_ENTER && mynamelen>0) {gamestate=GS_MAIN_MENU; menuopt=0; return;}
			else if (k>>8==KEY_BACKSPACE && mynamelen>0) myname[--mynamelen]=0;
			else if (k>>8==KEY_ESC) {
				if (mynamelen==0) exitgame(); else {gamestate=GS_MAIN_MENU; menuopt=0; return;}
			}
			else {
				k&=255;
				if (k>=32 && k<=126 && mynamelen<254) {
					myname[mynamelen++]=k; myname[mynamelen]=0;
					//Make sure this name will fit on the screen in the in-game display
					if (text_length((FONT *)dat[FONT_SMALL].dat,myname)>100) myname[--mynamelen]=0;
				}
			}
		}
	}
	if (gamestate==GS_MAIN_MENU) {
		while (keypressed()) {
			int k=readkey();
			if (k>>8==KEY_UP) menuopt=(menuopt+4)%5;
			if (k>>8==KEY_DOWN) menuopt=(menuopt+1)%5;
			if (k>>8==KEY_SPACE || k>>8==KEY_ENTER) {
				if (menuopt==0) {gamestate=GS_INSTRUCTIONS; return;}
				if (menuopt==1) {gamestate=GS_CONNECT; gamenamelen=0; gamename[0]=0; connectstep=0; return;}
				if (menuopt==2) {gamestate=GS_ENTERING_GAME_NAME; gamenamelen=0; gamename[0]=0; return;}
				if (menuopt==3) {gamestate=GS_ENTERING_NAME; return;}
				if (menuopt==4) exitgame();
			}
			if (k>>8==KEY_ESC) {gamestate=GS_ENTERING_NAME; return;}
		}
	}
	if (gamestate==GS_INSTRUCTIONS || gamestate==GS_CONNECT_ERROR) {
		while (keypressed()) {
			int k=readkey();
			if (k>>8==KEY_SPACE || k>>8==KEY_ENTER || k>>8==KEY_ESC) {gamestate=GS_MAIN_MENU; return;}
		}
	}
	if (gamestate==GS_ENTERING_GAME_NAME) {
		while (keypressed()) {
			int k=readkey();
			if (k>>8==KEY_ENTER && gamenamelen>0) {gamestate=GS_CONNECT; connectstep=0; return;}
			else if (k>>8==KEY_BACKSPACE && gamenamelen>0) gamename[--gamenamelen]=0;
			else if (k>>8==KEY_ESC) {gamestate=GS_MAIN_MENU; return;}
			else {
				k&=255;
				if (k>=32 && k<=126 && gamenamelen<254) {
					gamename[gamenamelen++]=k; gamename[gamenamelen]=0;
					//Make sure this name will fit on the screen in the in-game display TODO
					if (text_length((FONT *)dat[FONT_SMALL].dat,gamename)>100) gamename[--gamenamelen]=0;
				}
			}
		}
	}
	if (gamestate==GS_CONNECT) {
		//Check for abort
		while (keypressed()) {
			int k=readkey();
			if (k>>8==KEY_ESC) {
				if (con.socket!=NL_INVALID) {
					writeByte(con.out,con.outpos,C2S_CLOSING);
					writeByte(con.out,con.outpos,0);
					con.flushout(100);
					nlClose(con.socket);
					con.socket=NL_INVALID;
				}
				gamestate=GS_MAIN_MENU;
				return;
			}
		}
		//Process connect in steps
		if (connectstep==0) {
			clientDraw();
			blit(scrbuf,screen,0,0,0,0,640,480);
			if (!nlGetAddrFromName(host,&address)) {clientConnectError(); return;}
			if (!address.valid) {gamestate=GS_CONNECT_ERROR; strcpy(connecterror,"Host name does not resolve"); return;}
			connectstep=1;
			return;
		}
		if (connectstep==1) {
			con.socket=nlOpen(0,NL_RELIABLE);
			if (con.socket==NL_INVALID) {clientConnectError(); return;}
			if (!nlSetAddrPort(&address,port)) {clientConnectError(); return;}
			if (!nlConnect(con.socket,&address)) {clientConnectError(); return;}
			//Connected! Send our name, and the game name if any!
			con.inTimeout=con.outTimeout=0;
			con.inlen=con.outpos=0;
			writeByte(con.out,con.outpos,C2S_VERSION);
			writeByte(con.out,con.outpos,strlen(VERSION)+1);
			writeBlock(con.out,con.outpos,VERSION,strlen(VERSION)+1);
			writeByte(con.out,con.outpos,C2S_MY_NAME);
			writeByte(con.out,con.outpos,strlen(myname)+1);
			writeBlock(con.out,con.outpos,myname,strlen(myname)+1);
			writeByte(con.out,con.outpos,C2S_GAME_NAME);
			writeByte(con.out,con.outpos,strlen(gamename)+1);
			writeBlock(con.out,con.outpos,gamename,strlen(gamename)+1);
			connectstep=2;
			return;
		}
		if (connectstep==2) {
			con.flushout(100);//Wait for this not to go NL_CON_PENDING ...
			if (con.socket==NL_INVALID) {clientConnectError(); return;}
			if (con.outpos==0) connectstep=3;
			return;
		}
		//Once connectstep is 3, we use the packet reading loop below to go into the player list (and for many other state changes)
	}
	if (gamestate==GS_CHOOSING_CHARACTER) {
		while (keypressed()) {
			int k=readkey();
			if (k>>8==KEY_ESC) {
				writeByte(con.out,con.outpos,C2S_CLOSING);
				writeByte(con.out,con.outpos,0);
				con.flushout(100);
				nlClose(con.socket);
				con.socket=NL_INVALID;
				gamestate=GS_MAIN_MENU;
				return;
			}
			k&=255;
			if (k>='1' && k<='4') {
				int id=k-'1';
				if ((gameslots&(1<<id))==0) {
					writeByte(con.out,con.outpos,C2S_CHOOSE_CHARACTER+k-'1');
					writeByte(con.out,con.outpos,0);
				}
			}
		}
	}
	if (gamestate>=GS_SHOW_GAME_START) {
		//Keyboard input - chat, abort
		while (keypressed()) {
			int k=readkey();
			if (k>>8==KEY_ESC) {
				if (thegame!=NULL) {
					delete thegame; thegame=NULL;
					delete thefire; thefire=NULL;
				}
				writeByte(con.out,con.outpos,C2S_CLOSING);
				writeByte(con.out,con.outpos,0);
				con.flushout(100);
				nlClose(con.socket);
				con.socket=NL_INVALID;
				gamestate=GS_MAIN_MENU;
				return;
			}
			//End-of-game: play again - or - Player list: I am ready
			if (k>>8==KEY_ENTER && newchatlen==myminchatlen && (thegame==NULL || thegame->finished)) {
				if (thegame!=NULL) {
					delete thegame; thegame=NULL;
					delete thefire; thefire=NULL;
				}
				if (!playerready[myid]) {
					writeByte(con.out,con.outpos,C2S_READY);
					writeByte(con.out,con.outpos,0);
					playerready[myid]=true;
				}
				if (gamestate!=GS_PLAYER_LIST) {
					gamestate=GS_PLAYER_LIST;
					readySteadyMessage=0;
					return;
				}
			}
			//Chat
			else if (k>>8==KEY_ENTER && newchatlen>myminchatlen) {
				writeByte(con.out,con.outpos,C2S_CHAT);
				writeByte(con.out,con.outpos,newchatlen+1-myminchatlen);
				writeBlock(con.out,con.outpos,newchat+myminchatlen,newchatlen+1-myminchatlen);
				newchatlen=myminchatlen;
			}
			else if (k>>8==KEY_BACKSPACE && newchatlen>myminchatlen) newchat[--newchatlen]=0;
			else {
				if (newchatlen<254) {
					k&=255;
					if ((k==32 && newchatlen>myminchatlen) || (k>=33 && k<=126)) {
						newchat[newchatlen++]=k; newchat[newchatlen]=0;
						//Make sure this chat will fit on the screen
						if (text_length((FONT *)dat[FONT_SMALL].dat,newchat)>maxchatwidth) newchat[--newchatlen]=0;
					}
				}
			}
		}
		//Fire
		if (thegame!=NULL) thefire->update(thegame);
	}
	if (gamestate==GS_GAME && !thegame->finished) {
		//Check if we're getting behind on the inkeyqueue
		int oldlen=inlenhistory[inlenptr];
		int newlen=inkeylast-inkeyfirst;
		inlenhistory[inlenptr]=newlen;
		inlenptr=(inlenptr+1)&255;
		if (newlen<inminlen) inminlen=newlen;
		else if (oldlen==inminlen) {
			inminlen=511;
			for (int i=0; i<256; i++) {
				int v=inlenhistory[i];
				if (v<inminlen) inminlen=v;
			}
		}
		//Add input this frame to outkeyqueue - unless lagging too much
		//- don't do this if dead as reported by server
		//- this way, user appears to be able to control player until moment of death
		if (thegame->chr[myid].health>0 && outkeylast-outkeyfirst<511) {
			int keystate=0;
			if (key[KEY_LEFT]) keystate|=K_LEFT;
			if (key[KEY_RIGHT]) keystate|=K_RIGHT;
			if (key[KEY_UP]) keystate|=K_UP;
			if (key[KEY_DOWN]) keystate|=K_DOWN;
			if (key[KEY_SPACE] && newchatlen==myminchatlen) keystate|=K_SHOOT;
			outkeyqueue[outkeylast++&511]=keystate;
		}
		//If we have any incoming key states, use them
		if (inkeylast-inkeyfirst>0) useinkey();
		if (inminlen>5 && inkeylast-inkeyfirst>0) useinkey();
		//Send block of keystates in line with BURSTLEN
		if (outkeylast-outkeylastsent>=BURSTLEN) {
			writeByte(con.out,con.outpos,C2S_KEYSTATES);
			while (outkeylastsent<outkeylast) {
				int size=outkeylast-outkeylastsent;
				if (size>255) size=255;
				writeByte(con.out,con.outpos,size);
				for (int i=0; i<size; i++) {
					writeByte(con.out,con.outpos,outkeyqueue[outkeylastsent&511]);
					outkeylastsent++;
				}
			}
		}
		//Debugging!
		if (PRINT_KEYQUEUE_STATUS) {
			if ((logictick&8)==0) printf("Keyqueues: out at %d/511, in at %d/511\n",outkeylast-outkeyfirst,inkeylast-inkeyfirst);
		}
	}
	if (gamestate==GS_GAME && thegame->finished) {
		thegame->updateEnding(myid,myprojx,myprojy);
	}
	//Process connection
	if (con.socket!=NL_INVALID) {
		//Read packets from the server
		while (con.readin(100,S2C_START_GAME)) {
			int inpos=0;
			int type,len;
			readByte(con.in,inpos,type); type&=255;
			if (type==S2C_START_GAME) {readShort(con.in,inpos,len); len&=1023;}
			else {readByte(con.in,inpos,len); len&=255;}
			//printf("Got packet %d, len %d\n",type,len);
			if (type==KEEPALIVE) {
			}
			else if (type==S2C_SERVER_FULL) {
				nlClose(con.socket); con.socket=NL_INVALID;
				gamestate=GS_CONNECT_ERROR; strcpy(connecterror,"Sorry, server is full");
				return;
			}
			else if (type==S2C_WRONG_VERSION) {
				char version[256];
				readBlock(con.in,inpos,version,len);
				version[len>0?len-1:0]=0;
				nlClose(con.socket); con.socket=NL_INVALID;
				gamestate=GS_CONNECT_ERROR; sprintf(connecterror,"Server is \"%s\" but client is \"%s\"",version,VERSION);
				return;
			}
			else if (type==S2C_GAME_SLOTS) {
				if (len==1) {
					readByte(con.in,inpos,gameslots); gameslots&=15;
					gamestate=GS_CHOOSING_CHARACTER;
				}
			}
			else if (type>=S2C_MYID && type<S2C_MYID+4) {
				myid=type-S2C_MYID;
			}
			else if (type>=S2C_PLAYER_READY && type<S2C_PLAYER_WAITING+4) {
				if (gamestate==GS_CHOOSING_CHARACTER) {
					gamestate=GS_PLAYER_LIST;
					readySteadyMessage=0;
					playername[0][0]=playername[1][0]=playername[2][0]=playername[3][0]=0;
					playerready[0]=playerready[1]=playerready[2]=playerready[3]=false;
					for (int i=0; i<6; i++) thechat[i][0]=0;
					myminchatlen=newchatlen=sprintf(newchat,"<%s> ",myname);
				}
				if (type>=S2C_PLAYER_WAITING) {
					readBlock(con.in,inpos,playername[type-S2C_PLAYER_WAITING],len);
					playername[type-S2C_PLAYER_WAITING][len>0?len-1:0]=0;
					playerready[type-S2C_PLAYER_WAITING]=false;
				} else {
					readBlock(con.in,inpos,playername[type-S2C_PLAYER_READY],len);
					playername[type-S2C_PLAYER_READY][len>0?len-1:0]=0;
					playerready[type-S2C_PLAYER_READY]=true;
				}
			}
			else if (type>=S2C_PLAYER_LEFT && type<S2C_PLAYER_LEFT+4) {
				playername[type-S2C_PLAYER_LEFT][0]=0;
			}
			else if (type>=S2C_READY && type<=S2C_CANCEL_START) {
				readySteadyMessage=type;
			}
			else if (type==S2C_START_GAME) {
				for (int i=0; i<4; i++) playerready[i]=false;//Ready for next time
				gamestate=GS_GAME;
				thegame=new Game();
				for (int y=0; y<MAP_H; y++) readBlock(con.in,inpos,&thegame->map[y],MAP_W);
				for (int c=0; c<NUMCHARS; c++) {
					int active;
					readByte(con.in,inpos,active); active&=255;
					if (active!=0) {
						int x,y,movetype;
						readShort(con.in,inpos,x); x&=65535;
						readShort(con.in,inpos,y); y&=65535;
						readByte(con.in,inpos,movetype); movetype&=255;
						thegame->placePlayer(c,x,y,movetype);
					}
				}
				myprojx=thegame->chr[myid].x+thegame->chr[myid].width/2;
				myprojy=thegame->chr[myid].y+thegame->chr[myid].height/2;
				thefire=new Fire();
				initClientGame();
			}
			else if (type==S2C_KEYSTATES) {
				const int step=3+(NUMCHARS-4+1)/2;
				len=(len/step)*step;
				for (int i=0; i<len; i+=step) {
					int playerkeys,minerkeys=0,v;
					readByte(con.in,inpos,playerkeys); playerkeys&=255;
					readByte(con.in,inpos,v); playerkeys|=(v&255)<<8;
					readByte(con.in,inpos,v); playerkeys|=(v&255)<<16;
					if (NUMCHARS>4) {readByte(con.in,inpos,v); minerkeys|=v&255;}
					if (NUMCHARS>6) {readByte(con.in,inpos,v); minerkeys|=(v&255)<<8;}
					if (NUMCHARS>8) {readByte(con.in,inpos,v); minerkeys|=(v&255)<<16;}
					if (NUMCHARS>10) {readByte(con.in,inpos,v); minerkeys|=(v&255)<<24;}
					if (inkeylast-inkeyfirst>=511) useinkey();
					inplayerkeyqueue[inkeylast&511]=playerkeys;
					inminerkeyqueue[inkeylast&511]=minerkeys;
					inkeylast++;
				}
			}
			else if (type==S2C_CHAT) {
				for (int i=0; i<chatrows-1; i++) memcpy(thechat[i],thechat[i+1],sizeof(thechat[i]));
				readBlock(con.in,inpos,thechat[chatrows-1],len);
				thechat[chatrows-1][len>0?len-1:0]=0;
			}
			else {
				nlClose(con.socket); con.socket=NL_INVALID;
				gamestate=GS_CONNECT_ERROR; sprintf(connecterror,"Unrecognised packet %d",type);
				return;
			}
			con.inlen=0;
		}
		if (con.socket==NL_INVALID) {gamestate=GS_CONNECT_ERROR; strcpy(connecterror,"Connection failed"); return;}
		//Flush output to the server
		con.flushout(100);
		if (con.socket==NL_INVALID) {gamestate=GS_CONNECT_ERROR; strcpy(connecterror,"Connection failed"); return;}
	}
}

static void useinkey() {
	int playerkeys=inplayerkeyqueue[inkeyfirst&511];
	int minerkeys=inminerkeyqueue[inkeyfirst&511];
	inkeyfirst++;
	thegame->update(playerkeys,minerkeys,myid,myprojx,myprojy);
	if (outkeylast-outkeyfirst>0 && !((playerkeys>>(myid*PLAYER_KSHIFT))&K_SERVER_ENFORCED)) outkeyfirst++;
}

static void clientConnectError() {
	gamestate=GS_CONNECT_ERROR;
	if (nlGetError()==NL_SYSTEM_ERROR) strcpy(connecterror,nlGetSystemErrorStr(nlGetSystemError()));
	else strcpy(connecterror,nlGetErrorStr(nlGetError()));
}

static void exitgame() {
	nlShutdown();
	unload_datafile(dat);
	set_gfx_mode(GFX_TEXT,0,0,0,0);
	exit(EXIT_SUCCESS);
}
